#ifndef Memory_h_INCLUDED
#define Memory_h_INCLUDED

/*
 * !!!!	Memory.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 *
 */

/*
 * This module implements the memory allocator used by all the CDT containers and
 * data structures. It basically provides the "reap" interface proposed by
 * "Reconsidering Custom Memory Allocation" by E. D. Berger, B. G. Zorn, K. S.
 * McKinley. This is done on top of the system malloc().
 *
 * The "reap" interface provides a 'freeAll' method to free all the memory of a
 * Memory allocator at once, and also supports parent-child relationships between
 * Memory allocators so that deleting one recursively deletes its children.
 *
 * In addition to the "reap" interface, we enable memory pooling of small-sized
 * allocation requests. Pooling is controlled by a threshold supplied to Memory_new.
 * Currently pooled memory is only release when freeAll or Memory_delete is called.
 */

//
#define MemoryUsage_SMALLCOUNT 1024/CDT_ALIGN_MAX
#define MemoryUsage_LARGECOUNT 32-10			// log(2^32 / 1024).

//
struct MemoryUsageRecord_ {
  //@args
  uint32_t OPENED;		// Number of opened allocations in this size range.
  uint32_t ALLOCS;		// Total number of allocations in this size range.
};
typedef struct MemoryUsageRecord_ MemoryUsageRecord_, *MemoryUsageRecord;
typedef const struct MemoryUsageRecord_ *const_MemoryUsageRecord;
typedef struct MemoryUsageRecord_ * restrict_MemoryUsageRecord;

extern MemoryUsageRecord
MemoryUsageRecord_Ctor(MemoryUsageRecord this);

extern MemoryUsageRecord
MemoryUsageRecord_Copy(MemoryUsageRecord this, const_MemoryUsageRecord that);

extern void
MemoryUsageRecord_Dtor(MemoryUsageRecord this);

extern size_t
MemoryUsageRecord_Size(void);

#define MemoryUsageRecord_OPENED(this) ((this)->OPENED)
#define MemoryUsageRecord__OPENED(this) (&(this)->OPENED)
#define MemoryUsageRecord_ALLOCS(this) ((this)->ALLOCS)
#define MemoryUsageRecord__ALLOCS(this) (&(this)->ALLOCS)


/*
 * MemoryUsage --	State to track memory usage.
 */
struct MemoryUsage_ {
  //@args
  uint32_t INUSE;		// Current memory space in use.
  uint32_t MAXUSE;		// Maximum memory space used.
  uint32_t OPENED;		// Current number of allocations in all size ranges.
  uint32_t ALLOCS;		// Total number of allocations in all size ranges.
  MemoryUsageRecord_ SMALLRECORDS[MemoryUsage_SMALLCOUNT];
  MemoryUsageRecord_ LARGERECORDS[MemoryUsage_LARGECOUNT];
};
typedef struct MemoryUsage_ MemoryUsage_, *MemoryUsage;
typedef const struct MemoryUsage_ *const_MemoryUsage;
typedef struct MemoryUsage_ * restrict_MemoryUsage;

extern MemoryUsage
MemoryUsage_Ctor(MemoryUsage this);

extern MemoryUsage
MemoryUsage_Copy(MemoryUsage this, const_MemoryUsage that);

extern void
MemoryUsage_Dtor(MemoryUsage this);

extern size_t
MemoryUsage_Size(void);

#define MemoryUsage_INUSE(this) ((this)->INUSE)
#define MemoryUsage__INUSE(this) (&(this)->INUSE)
#define MemoryUsage_MAXUSE(this) ((this)->MAXUSE)
#define MemoryUsage__MAXUSE(this) (&(this)->MAXUSE)
#define MemoryUsage_OPENED(this) ((this)->OPENED)
#define MemoryUsage__OPENED(this) (&(this)->OPENED)
#define MemoryUsage_ALLOCS(this) ((this)->ALLOCS)
#define MemoryUsage__ALLOCS(this) (&(this)->ALLOCS)
#define MemoryUsage_SMALLRECORDS(this) ((this)->SMALLRECORDS)
#define MemoryUsage__SMALLRECORDS(this) (&(this)->SMALLRECORDS)
#define MemoryUsage_LARGERECORDS(this) ((this)->LARGERECORDS)
#define MemoryUsage__LARGERECORDS(this) (&(this)->LARGERECORDS)


/*
 * MemoryUsage_pretty --	Pretty-print this MemoryUsage.
 */
bool
MemoryUsage_pretty(const_MemoryUsage this, FILE *file);

/*
 * MemoryUsage_alloc --	Trace usage of =malloc=.
 */
void
MemoryUsage_alloc(MemoryUsage this, uint32_t size, bool isPool);

/*
 * MemoryUsage_free --	Trace usage of =free=.
 */
void
MemoryUsage_free(MemoryUsage this, uint32_t size, bool isPool);

//
extern MemoryUsage Memory_Usage;

/*
 * MemoryListItem --	Fields needed by the items of a MemoryList.
 */
struct MemoryListItem_ {
  //@args
  struct MemoryListItem_ *NEXT;
  struct MemoryListItem_ *PREV;
};
typedef struct MemoryListItem_ MemoryListItem_, *MemoryListItem;
typedef const struct MemoryListItem_ *const_MemoryListItem;
typedef struct MemoryListItem_ * restrict_MemoryListItem;

extern MemoryListItem
MemoryListItem_Ctor(MemoryListItem this);

extern MemoryListItem
MemoryListItem_Copy(MemoryListItem this, const_MemoryListItem that);

extern void
MemoryListItem_Dtor(MemoryListItem this);

extern size_t
MemoryListItem_Size(void);

#define MemoryListItem_NEXT(this) ((this)->NEXT)
#define MemoryListItem__NEXT(this) (&(this)->NEXT)
#define MemoryListItem_PREV(this) ((this)->PREV)
#define MemoryListItem__PREV(this) (&(this)->PREV)


/*
 * MemoryList --	Manage a list of items derived from MemoryListItem.
 */
struct MemoryList_ {
  //@args
  struct MemoryListItem_ *FIRST;
  struct MemoryListItem_ *LAST;
};
typedef struct MemoryList_ MemoryList_, *MemoryList;
typedef const struct MemoryList_ *const_MemoryList;
typedef struct MemoryList_ * restrict_MemoryList;

extern MemoryList
MemoryList_Ctor(MemoryList this);

extern MemoryList
MemoryList_Copy(MemoryList this, const_MemoryList that);

extern void
MemoryList_Dtor(MemoryList this);

extern size_t
MemoryList_Size(void);

#define MemoryList_FIRST(this) ((this)->FIRST)
#define MemoryList__FIRST(this) (&(this)->FIRST)
#define MemoryList_LAST(this) ((this)->LAST)
#define MemoryList__LAST(this) (&(this)->LAST)


/*
 * MemoryList_first -- For use by MemoryList_FOREACH_MemoryListItem.
 */
static inline MemoryListItem
MemoryList_first(MemoryList this)
{
  return MemoryList_FIRST(this);
}

//
bool
MemoryList_check(MemoryList this, MemoryListItem item);

/*
 * MemoryList_FOREACH_MemoryListItem --	Iterator.
 */
#define MemoryList_FOREACH_MemoryListItem(this, item) { \
  MemoryListItem MemoryList_NEXT = NULL; \
  MemoryListItem item = MemoryList_first(this); \
  for (; item != NULL; item = MemoryList_NEXT) { \
    MemoryList_NEXT = MemoryListItem_NEXT(item);
#define MemoryList_ENDEACH_MemoryListItem \
  } \
}

/*
 * MemoryList_append --	Append item to this MemoryList.
 */
void
MemoryList_append(MemoryList this, MemoryListItem item);

/*
 * MemoryList_remove --	Remove item from this MemoryList.
 */
void
MemoryList_remove(MemoryList this, MemoryListItem item);

/*
 * MemoryTableEntry --	Entry in a MemoryTable.
 */
struct MemoryTableEntry_ {
  //@args	void *start, size_t useSize, int level,
  //@args	struct MemoryTableEntry_ *leaf
  size_t USESIZE;			// Size of memory available for use.
  uint16_t LEVEL;			// Level to keep the binary search tree balanced.
  uint16_t REFC;			// Reference counter of this MemoryTableEntry_.
  struct MemoryTableEntry_ *LEFT;	// The left kid in the binary search tree.
  struct MemoryTableEntry_ *RIGHT;	// The right kid in the binary search tree.
  char *START;				// The start memory address.
  //@access PAST	(MemoryTableEntry_START(this) + MemoryTableEntry_USESIZE(this))
  struct MemoryBlock_ *NEXTBLOCK;	// Next MemoryBlock in the Memory_RECYCLE list.
  //@access BLOCK	(MemoryBlock)((MemoryListItem)(this) - 1)
};
typedef struct MemoryTableEntry_ MemoryTableEntry_, *MemoryTableEntry;
typedef const struct MemoryTableEntry_ *const_MemoryTableEntry;
typedef struct MemoryTableEntry_ * restrict_MemoryTableEntry;

extern MemoryTableEntry
MemoryTableEntry_Ctor(MemoryTableEntry this, void *start, size_t useSize, int level, struct MemoryTableEntry_ *leaf);

extern MemoryTableEntry
MemoryTableEntry_Copy(MemoryTableEntry this, const_MemoryTableEntry that);

extern void
MemoryTableEntry_Dtor(MemoryTableEntry this);

extern size_t
MemoryTableEntry_Size(void *start, size_t useSize, int level, struct MemoryTableEntry_ *leaf);

#define MemoryTableEntry_USESIZE(this) ((this)->USESIZE)
#define MemoryTableEntry__USESIZE(this) (&(this)->USESIZE)
#define MemoryTableEntry_LEVEL(this) ((this)->LEVEL)
#define MemoryTableEntry__LEVEL(this) (&(this)->LEVEL)
#define MemoryTableEntry_REFC(this) ((this)->REFC)
#define MemoryTableEntry__REFC(this) (&(this)->REFC)
#define MemoryTableEntry_LEFT(this) ((this)->LEFT)
#define MemoryTableEntry__LEFT(this) (&(this)->LEFT)
#define MemoryTableEntry_RIGHT(this) ((this)->RIGHT)
#define MemoryTableEntry__RIGHT(this) (&(this)->RIGHT)
#define MemoryTableEntry_START(this) ((this)->START)
#define MemoryTableEntry__START(this) (&(this)->START)
#define MemoryTableEntry_NEXTBLOCK(this) ((this)->NEXTBLOCK)
#define MemoryTableEntry__NEXTBLOCK(this) (&(this)->NEXTBLOCK)
#define MemoryTableEntry_BLOCK(this) ((MemoryBlock)((MemoryListItem)(this) - 1))
#define MemoryTableEntry_PAST(this) ((MemoryTableEntry_START(this) + MemoryTableEntry_USESIZE(this)))


/*
 * MemoryTable --	Maps addresses to MemoryBlock(s).
 *
 * Currently the MemoryTable only allows search and insert operations.
 *
 * Uses a balanced search tree as described in "Balanced Search Trees Made Simple"
 * by Arne Andersson (http://user.it.uu.se/~arnea/research.html).
 */
struct MemoryTable_ {
  //@args	void *memory
  MemoryTableEntry_ LEAF_;
  //@access LEAF	MemoryTable__LEAF_(this)
  MemoryTableEntry ROOT;
};
typedef struct MemoryTable_ MemoryTable_, *MemoryTable;
typedef const struct MemoryTable_ *const_MemoryTable;
typedef struct MemoryTable_ * restrict_MemoryTable;

extern MemoryTable
MemoryTable_Ctor(MemoryTable this, void *memory);

extern MemoryTable
MemoryTable_Copy(MemoryTable this, const_MemoryTable that);

extern void
MemoryTable_Dtor(MemoryTable this);

extern size_t
MemoryTable_Size(void *memory);

#define MemoryTable_LEAF_(this) ((this)->LEAF_)
#define MemoryTable__LEAF_(this) (&(this)->LEAF_)
#define MemoryTable_ROOT(this) ((this)->ROOT)
#define MemoryTable__ROOT(this) (&(this)->ROOT)
#define MemoryTable_LEAF(this) (MemoryTable__LEAF_(this))


/*
 * MemoryTable_search --	Search this MemoryTable for the MemoryBlock containing =address=.
 */
struct MemoryBlock_ *
MemoryTable_search(const_MemoryTable this, const void *address);

/*
 * MemoryBlock --	Block of memory.
 */
struct MemoryBlock_ {
  //@args	size_t useSize, MemoryTableEntry leaf
  MemoryListItem_ ITEM_;
  //@access ITEM	MemoryBlock__ITEM_(this)
  //@access NEXT	(MemoryBlock)MemoryListItem_NEXT(MemoryBlock_ITEM(this))
  //@access PREV	(MemoryBlock)MemoryListItem_PREV(MemoryBlock_ITEM(this))
  MemoryTableEntry_ ENTRY_;
  //@access ENTRY	MemoryBlock__ENTRY_(this)
  //@access NEXTBLOCK	MemoryTableEntry_NEXTBLOCK(MemoryBlock_ENTRY(this))
  //@mutate NEXTBLOCK	MemoryTableEntry__NEXTBLOCK(MemoryBlock_ENTRY(this))
  //@access REALSIZE	(sizeof(MemoryBlock_) + MemoryBlock_USESIZE(this))
  //@access USESIZE	MemoryTableEntry_USESIZE(MemoryBlock_ENTRY(this))
  //@mutate USESIZE	MemoryTableEntry__USESIZE(MemoryBlock_ENTRY(this))
  //@access START	MemoryTableEntry_START(MemoryBlock_ENTRY(this))
  //@mutate START	MemoryTableEntry__START(MemoryBlock_ENTRY(this))
  //@access PAST	MemoryTableEntry_PAST(MemoryBlock_ENTRY(this))
  //@access REFC	MemoryTableEntry_REFC(MemoryBlock_ENTRY(this))
  //@mutate REFC	MemoryTableEntry__REFC(MemoryBlock_ENTRY(this))
};
typedef struct MemoryBlock_ MemoryBlock_, *MemoryBlock;
typedef const struct MemoryBlock_ *const_MemoryBlock;
typedef struct MemoryBlock_ * restrict_MemoryBlock;

extern MemoryBlock
MemoryBlock_Ctor(MemoryBlock this, size_t useSize, MemoryTableEntry leaf);

extern MemoryBlock
MemoryBlock_Copy(MemoryBlock this, const_MemoryBlock that);

extern void
MemoryBlock_Dtor(MemoryBlock this);

extern size_t
MemoryBlock_Size(size_t useSize, MemoryTableEntry leaf);

#define MemoryBlock_ITEM_(this) ((this)->ITEM_)
#define MemoryBlock__ITEM_(this) (&(this)->ITEM_)
#define MemoryBlock_ENTRY_(this) ((this)->ENTRY_)
#define MemoryBlock__ENTRY_(this) (&(this)->ENTRY_)
#define MemoryBlock_NEXT(this) ((MemoryBlock)MemoryListItem_NEXT(MemoryBlock_ITEM(this)))
#define MemoryBlock_REFC(this) (MemoryTableEntry_REFC(MemoryBlock_ENTRY(this)))
#define MemoryBlock_ENTRY(this) (MemoryBlock__ENTRY_(this))
#define MemoryBlock_REALSIZE(this) ((sizeof(MemoryBlock_) + MemoryBlock_USESIZE(this)))
#define MemoryBlock_NEXTBLOCK(this) (MemoryTableEntry_NEXTBLOCK(MemoryBlock_ENTRY(this)))
#define MemoryBlock_ITEM(this) (MemoryBlock__ITEM_(this))
#define MemoryBlock_USESIZE(this) (MemoryTableEntry_USESIZE(MemoryBlock_ENTRY(this)))
#define MemoryBlock_PAST(this) (MemoryTableEntry_PAST(MemoryBlock_ENTRY(this)))
#define MemoryBlock_START(this) (MemoryTableEntry_START(MemoryBlock_ENTRY(this)))
#define MemoryBlock_PREV(this) ((MemoryBlock)MemoryListItem_PREV(MemoryBlock_ITEM(this)))
#define MemoryBlock__NEXTBLOCK(this) (MemoryTableEntry__NEXTBLOCK(MemoryBlock_ENTRY(this)))
#define MemoryBlock__START(this) (MemoryTableEntry__START(MemoryBlock_ENTRY(this)))
#define MemoryBlock__REFC(this) (MemoryTableEntry__REFC(MemoryBlock_ENTRY(this)))
#define MemoryBlock__USESIZE(this) (MemoryTableEntry__USESIZE(MemoryBlock_ENTRY(this)))


/*
 * Memory -- Memory allocator.
 */
struct Memory_ {
  //@args	Memory parent, size_t pooled
  MemoryListItem_ ITEM_;	// Used to embed this Memory in a MemoryList.
  //@access ITEM	Memory__ITEM_(this)
  //@access NEXT	(Memory)MemoryListItem_NEXT(Memory_ITEM(this))
  //@access PREV	(Memory)MemoryListItem_PREV(Memory_ITEM(this))
  struct Memory_ *PARENT;	// Parent in the Memory parent-child relationships.
  //@access isRoot	(Memory_PARENT(this) == NULL)
  size_t POOLED;		// Maximum alloc size for request to be pooled.
  size_t TOTALSIZE;		// Total allocated size (for statistics).
  size_t TOTALCOUNT;		// Total allocated count (for statistics).
  MemoryBlock POOLBLOCK;	// Currently active pooling block.
  char *POOLFIRST;		// First available address in pooling block.
  char *POOLPAST;		// Past available address in pooling block.
  MemoryBlock RECYCLE;		// Start of list of MemoryBlock(s) to recycle.
  MemoryList_ CHILDREN_;	// Children of this Memory allocator.
  //@access CHILDREN	Memory__CHILDREN_(this)
  MemoryList_ BLOCKS_;		// MemoryBlock(s) managed by this Memory allocator.
  //@access BLOCKS	Memory__BLOCKS_(this)
  //@access TABLE	(MemoryTable)(this + 1)
};
typedef struct Memory_ Memory_, *Memory;
typedef const struct Memory_ *const_Memory;
typedef struct Memory_ * restrict_Memory;

extern Memory
Memory_Ctor(Memory this, Memory parent, size_t pooled);

extern Memory
Memory_Copy(Memory this, const_Memory that);

extern void
Memory_Dtor(Memory this);

extern size_t
Memory_Size(Memory parent, size_t pooled);

#define Memory_ITEM_(this) ((this)->ITEM_)
#define Memory__ITEM_(this) (&(this)->ITEM_)
#define Memory_PARENT(this) ((this)->PARENT)
#define Memory__PARENT(this) (&(this)->PARENT)
#define Memory_POOLED(this) ((this)->POOLED)
#define Memory__POOLED(this) (&(this)->POOLED)
#define Memory_TOTALSIZE(this) ((this)->TOTALSIZE)
#define Memory__TOTALSIZE(this) (&(this)->TOTALSIZE)
#define Memory_TOTALCOUNT(this) ((this)->TOTALCOUNT)
#define Memory__TOTALCOUNT(this) (&(this)->TOTALCOUNT)
#define Memory_POOLBLOCK(this) ((this)->POOLBLOCK)
#define Memory__POOLBLOCK(this) (&(this)->POOLBLOCK)
#define Memory_POOLFIRST(this) ((this)->POOLFIRST)
#define Memory__POOLFIRST(this) (&(this)->POOLFIRST)
#define Memory_POOLPAST(this) ((this)->POOLPAST)
#define Memory__POOLPAST(this) (&(this)->POOLPAST)
#define Memory_RECYCLE(this) ((this)->RECYCLE)
#define Memory__RECYCLE(this) (&(this)->RECYCLE)
#define Memory_CHILDREN_(this) ((this)->CHILDREN_)
#define Memory__CHILDREN_(this) (&(this)->CHILDREN_)
#define Memory_BLOCKS_(this) ((this)->BLOCKS_)
#define Memory__BLOCKS_(this) (&(this)->BLOCKS_)
#define Memory_PREV(this) ((Memory)MemoryListItem_PREV(Memory_ITEM(this)))
#define Memory_ITEM(this) (Memory__ITEM_(this))
#define Memory_CHILDREN(this) (Memory__CHILDREN_(this))
#define Memory_TABLE(this) ((MemoryTable)(this + 1))
#define Memory_BLOCKS(this) (Memory__BLOCKS_(this))
#define Memory_isRoot(this) ((Memory_PARENT(this) == NULL))
#define Memory_NEXT(this) ((Memory)MemoryListItem_NEXT(Memory_ITEM(this)))


/*
 * Memory_new --	Make a new Memory allocator.
 *
 * =parent=:	The parent memory allocator.
 * =pooled=:	The maximum allocation size pooled by this Memory allocator.
 *
 * Memory allocators are organized in tree thanks to the =parent= parameter. When a
 * Memory allocator is deleted or does freeAll, it recursively deletes all its children.
 *
 * The =pooled= parameter is the threshold on allocation size beyond which the
 * =Memory_alloc= and =Memory_free= directly forward to =malloc= and =free=. For
 * allocations sizes no greater than =pooled=, the requests are pooled by the
 * Memory allocator.
 */
Memory
Memory_new(Memory parent, size_t pooled);

/*
 * Memory_delete --	Delete this Memory allocator.
 */
Memory
Memory_delete(Memory this);

/*
 * Memory_children --	For use by Memory_CHILDREN_FOREACH_Memory.
 */
static inline MemoryList
Memory_children(Memory this)
{
  return Memory_CHILDREN(this);
}

/*
 * Memory_CHILDREN_FOREACH_Memory --	Iterator for Memory_CHILDREN.
 */
#define Memory_CHILDREN_FOREACH_Memory(this, child) { \
  MemoryList Memory_CHILDREN = Memory_children(this); \
  MemoryList_FOREACH_MemoryListItem(Memory_CHILDREN, item) { \
    Memory child = (Memory)item;
#define Memory_CHILDREN_ENDEACH_Memory \
  } MemoryList_ENDEACH_MemoryListItem; \
}

/*
 * Memory_blocks --	For use by Memory_FOREACH_MemoryBlock.
 */
static inline MemoryList
Memory_blocks(Memory this)
{
  return Memory_BLOCKS(this);
}

/*
 * Memory_FOREACH_MemoryBlock --	Iterator for Memory_BLOCKS.
 */
#define Memory_FOREACH_MemoryBlock(this, block) { \
  MemoryList Memory_BLOCKS = Memory_blocks(this); \
  MemoryList_FOREACH_MemoryListItem(Memory_BLOCKS, item) { \
    MemoryBlock block = (MemoryBlock)item;
#define Memory_ENDEACH_MemoryBlock \
  } MemoryList_ENDEACH_MemoryListItem; \
}

/*
 * Memory_POOLFACTOR -- Multiplied by Memory_POOLED to compute pooling block use size.
 */
#ifndef Memory_POOLFACTOR
#define Memory_POOLFACTOR 8
#endif//Memory_POOLFACTOR

/*
 * Memory_alloc_ --	Specialized =Memory_alloc= for non-zero aligned =alignedSize=.
 */
void *
Memory_alloc_(Memory this, size_t alignedSize);

/*
 * Memory_alloc --	Allocate memory in its own MemoryBlock or in a pool.
 *
 * =size=:	The requested allocation size in bytes.
 * Return:	Allocated address that holds =size= bytes.
 * Ensure:	Returned address is non-NULL.
 */
static inline void *
Memory_alloc(Memory this, size_t size)
{
  Except_CHECK(this != NULL);
  if (size > 0) {
    size_t alignedSize = CDT_ALIGN_NEXT(size);
    void *alloc = Memory_alloc_(this, alignedSize);
    return alloc;
  }
  return NULL;
}

/*
 * Memory_free_ --	Specialized =Memory_free= for non-NULL =address=.
 */
void
Memory_free_(Memory this, void *address);

/*
 * Memory_free --	Free the memory at given =address= unless it is pooled.
 *
 * Pooled memory is not returned to the system until =this= Memory is deleted.
 * However, reference counting of pooling memory blocks is maintained so when all
 * the allocations of such a block are freed it is recycled.
 */
static inline void *
Memory_free(Memory this, void *address)
{
  Except_CHECK(this != NULL);
  if (address != NULL) {
    Except_CHECK(CDT_ALIGNED(address));
    Memory_free_(this, address);
  }
  return NULL;
}

/*
 * Memory_freeAll --	Free all the MemoryBlock(s) and children of this Memory allocator.
 */
void
Memory_freeAll(Memory this);

//
void *
Memory_realloc_(Memory this, void* address, size_t useSize);

/*
 * Memory_realloc --	Reallocate data allocated by this Memory allocator.
 */
void *
Memory_realloc(Memory this, void* address, size_t size);

//
extern Memory Memory_Root;

//
uint32_t
Memory_usageMax(void);

//
uint32_t
Memory_usageCurrent(void);


#endif//Memory_h_INCLUDED
