#ifndef Network__h_INCLUDED
#define Network__h_INCLUDED

//
#include "CAL/Network.h"

//
static inline int32_t
NetArc_reduced(NetArc this)
{
  NetNode t_node = NetArc_TAILNODE(this), h_node = NetArc_HEADNODE(this);
  return NetArc_DELAY(this) + NetNode_DATE(t_node) - NetNode_DATE(h_node);
}

/*
 * Network_QUEUED_isEmpty --	True if this Network_QUEUED is empty.
 */
static inline bool
Network_QUEUED_isEmpty(Network this)
{
  return IQueue_isEmpty(Network_QUEUED(this));
}

/*
 * Network_QUEUED_putArc --	Put a NetArc on this Network_QUEUED.
 */
static inline void
Network_QUEUED_putArc(Network this, NetArc netarc)
{
  if (!NetArc_isQueued(netarc)) {
    *NetArc__FLAGS(netarc) |= NetArcFlag_Queued;
    IQueue_PUT(Network_QUEUED(this), NetArc, netarc);
  }
}

/*
 * Network_QUEUED_putNode --	Put the incident arcs of a NetNode on this Network_QUEUED.
 */
static inline void
Network_QUEUED_putNode(Network this, NetNode netnode)
{
  NetNode_FORENTER_NetArc(netnode, netarc) {
    Network_QUEUED_putArc(this, netarc);
  } NetNode_ENDENTER_NetArc;
  NetNode_FORLEAVE_NetArc(netnode, netarc) {
    Network_QUEUED_putArc(this, netarc);
  } NetNode_ENDLEAVE_NetArc;
}

/*
 * Network_QUEUED_topArc --	The top NetArc of this Network_QUEUED.
 */
static inline NetArc
Network_QUEUED_topArc(Network this)
{
  return IQueue_TOP(Network_QUEUED(this), NetArc);
}

/*
 * Network_QUEUED_bottomArc --	The bottom NetArc of this Network_QUEUED.
 */
static inline NetArc
Network_QUEUED_bottomArc(Network this)
{
  return IQueue_BOT(Network_QUEUED(this), NetArc);
}

/*
 * Network_QUEUED_popArc --	Pop the top NetArc from this Network_QUEUED.
 */
static inline NetArc
Network_QUEUED_popArc(Network this)
{
  NetArc netarc = NULL;
  IQueue_POP(Network_QUEUED(this), NetArc, netarc);
  Except_CHECK(NetArc_isQueued(netarc));
  *NetArc__FLAGS(netarc) &= ~NetArcFlag_Queued;
  return netarc;
}

/*
 * Network_QUEUED_dropArc --	Drop the bottom NetArc from this Network_QUEUED.
 */
static inline NetArc
Network_QUEUED_dropArc(Network this)
{
  NetArc netarc = NULL;
  IQueue_DROP(Network_QUEUED(this), NetArc, netarc);
  Except_CHECK(NetArc_isQueued(netarc));
  *NetArc__FLAGS(netarc) &= ~NetArcFlag_Queued;
  return netarc;
}

/*
 * Network_doGST --	Grow Spanning Tree.
 *
 * Function used to add node =l= to the spanning tree from node =k=, taking arc
 * =q= as the new tree edge. The tree indices (NEXT, PREV, DEPTH) and the node
 * potentials (DATE) are updated. (No longer used.)
 */
static inline void
Network_doGST(Network this, NetArc q, NetNode k, NetNode l)
{
  *NetNode__EDGE(l) = q;
  *NetNode__DEPTH(l) = NetNode_DEPTH(k) + 1;
  *NetNode__DATE(l) = NetNode_DATE(k) + NetArc_DELAY(q);
  {
    NetNode n = NetNode_NEXT(k);
    *NetNode__NEXT(l) = n, *NetNode__PREV(n) = l;
    *NetNode__NEXT(k) = l, *NetNode__PREV(l) = k;
  }
}

/*
 * Network_doRST --	Reorder Spanning Tree for a network simplex pivot.
 *
 * Function used to enter arc =q= as the tree edge between node =k= and node @p
 * l in the spanning tree. The leaving arc is node =j= spanning tree edge.
 */
void
Network_doRST(Network this, bool queue);

/*
 * Network_doCDC --	Critical Directed Cycles.
 *
 * Find a critical directed cycle. The first step is
 * to identify the directed cycle by going up the spanning tree from =k=. If
 * there is a directed cycle, it is infeasible in case its delay > 0. The
 * CRITICAL call-back is then called to modify the arc delays.
 */
bool
Network_doCDC(Network this, NetArc q, NetNode k, NetNode l);

/*
 * Network_doFAQ --	Fill Arc Queue.
 *
 * Fill the Network_QUEUED with the arcs that violate Bellman's inequalities.
 */
bool
Network_doFAQ(Network this, bool refill);

/*
 * Network_doUEF --	Update Edge Flows for a network simplex pivot.
 *
 * Given the entering arc =q= and the leaving edge =p=, the flows are updated.
 * This function assumes that the spanning tree is not yet reordered by @c
 * Network_doRST.
 */
void
Network_doUEF(Network this, bool queue);

/*
 * Network_dualFLE --	Find Leaving Edge of the dual network simplex algorithm.
 *
 * Ideally, this is an edge with the minimum negative flow. This is what @c
 * Network_dualFLE actually does, but this implementation is slow. TODO:
 * implement a cached scheme.
 */
bool
Network_dualFLE(Network this);

/*
 * Network_dualFEA --	Find Entering Arc of the dual network simplex algorithm.
 *
 * We first identify the cut-set of arcs with the opposite direction. In this
 * set, the entering arc has the maximum (non-positive) reduced delay. To our
 * knowledge, the strongly feasible basis property cannot be maintained.
 */
void
Network_dualFEA(Network this);

/*
 * Network_primalFLE --	Find Leaving Edge of the primal network simplex algorithm.
 *
 * This algorithm maintains strongly feasible bases, as defined by Cunningham in
 * "A Network Simplex Method", Mathematical Programming 11 (1976): tree edges
 * with zero flow are directed away from the root. This is achieved by starting
 * from the apex, and selecting the first blocking arc in the cycle following the
 * direction of the entering arc. Please note that Ahuja, Magnanti Orlin in
 * "Network Flows", Prentice-Hall 1993, use the opposite definition of strongly
 * feasible bases, so they select the last blocking arc instead.
 */
void
Network_primalFLE(Network this);

/*
 * Network_primalFEA --	Find Entering Arc of the primal network simplex algorithm.
 *
 * Ideally, this is an arc with the maximum reduced delay. In order to save
 * computations, we apply a cached major / minor iteration scheme. This relies on
 * the Network_QUEUED where the arcs that are likely to violate Bellman's
 * inequalities are put by @c Network_doUEF and @c Network_doRST.
 */
bool
Network_primalFEA(Network this, bool refill);

/*
 * Network_INIT --	Initialize the Network module.
 */
#ifndef CDT_NTRACE
void
Network_INIT(void);
#else//CDT_NTRACE
#define Network_INIT()
#endif//CDT_NTRACE

/*
  Network_FINI -- Finalize the Network module.
 */
#ifndef CDT_NTRACE
void
Network_FINI(void);
#else//CDT_NTRACE
#define Network_FINI()
#endif//CDT_NTRACE


#endif//Network__h_INCLUDED
