#ifndef Scoped_h_INCLUDED
#define Scoped_h_INCLUDED

/*
 * !!!!	Scoped.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
typedef const void *ScopedKey;
typedef void *ScopedValue;
typedef void (*ScopedValueRelease)(ScopedValue);

/*
  ScopedEntry -- Object inlined below ScopedEntry.
 */
struct ScopedEntry_ {
  //@args	void *symbol, int32_t scope
  void *SYMBOL;
  int32_t SCOPE;
};
typedef struct ScopedEntry_ ScopedEntry_, *ScopedEntry;
typedef const struct ScopedEntry_ *const_ScopedEntry;
typedef struct ScopedEntry_ * restrict_ScopedEntry;

extern ScopedEntry
ScopedEntry_Ctor(ScopedEntry this, void *symbol, int32_t scope);

extern ScopedEntry
ScopedEntry_Copy(ScopedEntry this, const_ScopedEntry that);

extern void
ScopedEntry_Dtor(ScopedEntry this);

extern size_t
ScopedEntry_Size(void *symbol, int32_t scope);

#define ScopedEntry_SYMBOL(this) ((this)->SYMBOL)
#define ScopedEntry__SYMBOL(this) (&(this)->SYMBOL)
#define ScopedEntry_SCOPE(this) ((this)->SCOPE)
#define ScopedEntry__SCOPE(this) (&(this)->SCOPE)


//
struct Scoped_ {
  //@args	Memory memory, bool copies, int32_t estimate
  StringTable_ STRINGS_;        // Local StringTable.
  //@access STRINGS     Scoped__STRINGS_(this)
  HTable_ HTABLE_;		// Maps symbol key to ScopedEntry.
  //@access HTABLE	Scoped__HTABLE_(this)
  //@access MEMORY	HTable_MEMORY(Scoped_HTABLE(this))
  IDList_ IDLIST_;		// Where the symbols are inlined.
  //@access IDLIST	Scoped__IDLIST_(this)
  //@access COUNT	IDList_count(Scoped_IDLIST(this))
};
typedef struct Scoped_ Scoped_, *Scoped;
typedef const struct Scoped_ *const_Scoped;
typedef struct Scoped_ * restrict_Scoped;

extern Scoped
Scoped_Ctor(Scoped this, Memory memory, bool copies, int32_t estimate);

extern Scoped
Scoped_Copy(Scoped this, const_Scoped that);

extern void
Scoped_Dtor(Scoped this);

extern size_t
Scoped_Size(Memory memory, bool copies, int32_t estimate);

#define Scoped_STRINGS_(this) ((this)->STRINGS_)
#define Scoped__STRINGS_(this) (&(this)->STRINGS_)
#define Scoped_HTABLE_(this) ((this)->HTABLE_)
#define Scoped__HTABLE_(this) (&(this)->HTABLE_)
#define Scoped_IDLIST_(this) ((this)->IDLIST_)
#define Scoped__IDLIST_(this) (&(this)->IDLIST_)
#define Scoped_IDLIST(this) (Scoped__IDLIST_(this))
#define Scoped_HTABLE(this) (Scoped__HTABLE_(this))
#define Scoped_MEMORY(this) (HTable_MEMORY(Scoped_HTABLE(this)))
#define Scoped_COUNT(this) (IDList_count(Scoped_IDLIST(this)))
#define Scoped_STRINGS(this) (Scoped__STRINGS_(this))


/*
 * Scoped_empty --	Empty this Scoped.
 */
void
Scoped_empty(Scoped this, ScopedValueRelease release);

/*
 * Scoped_memory --	This Scoped memory.
 */
static inline Memory
Scoped_memory(const_Scoped this)
{
  return Scoped_MEMORY(this);
}

/*
 * Scoped_idlist --	For use in Scoped_FOREACH.
 */
static inline const_IDList
Scoped_idlist(const_Scoped this)
{
  return Scoped_IDLIST(this);
}

/*
 * Scoped_count --	Count entries in the Scoped.
 * Return:	The count of entries in the Scoped.
 */
static inline int32_t
Scoped_count(const_Scoped this)
{
  return Scoped_COUNT(this);
}

/*
 * Scoped_search --	Search the symbol associated to =key=.
 * Return:	The corresponding ScopedEntry if =key= is in Scoped, else NULL.
 */
void *
Scoped_search(const_Scoped this, ScopedKey key, int32_t *_scope);

/*
 * Scoped_lookup --	Lookup the symbol at =key=, insert if not already there.
 */
void *
Scoped_lookup(Scoped this, ScopedKey key, size_t symbolSize, int32_t scope);

/*
 * Scoped_insert --	Insert the symbol at =key= if not already there.
 */
void *
Scoped_insert(Scoped this, ScopedKey key, size_t symbolSize, int32_t scope);

/*
 * Scoped_install --	Insert a symbol even if another has same key but not same scope.
 */
void *
Scoped_install(Scoped this, ScopedKey key, size_t symbolSize, int32_t scope);

/*
 * Scoped_remove --	Remove =key= and the corresponding symbol.
 * Return:	=true= if =key= was in Scoped, else =false=.
 */
bool
Scoped_remove(Scoped this, ScopedKey key, int32_t scope);

/*
 * Scoped_purgeScope --	Purge symbols of higher scopes.
 */
void
Scoped_purgeScope(Scoped this);

/*
 * Scoped_FOREACH --	Iterates from first to last Scoped entry.
 */
#define Scoped_FOREACH(this, Type, iter) \
  IDList_FOREACH(Scoped_idlist(this), Type, iter)
#define Scoped_ENDEACH \
  IDList_ENDEACH;

/*
 * Scoped_FORBACK --	Iterates from last to first Scoped entry.
 */
#define Scoped_FORBACK(this, Type, iter) \
  IDList_FORBACK(Scoped_idlist(this), Type, iter)
#define Scoped_ENDBACK \
  IDList_ENDBACK;


#endif//Scoped_h_INCLUDED
