#ifndef Equiv_c_INCLUDED
#define Equiv_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
/*
 *
 * Implementation of Equivalence class based on UNION-FIND
 *
 * sparse_: This array is used to avoid prenumbering all Temporaries
 *          in the program. A Temporary has a valid INDEX if and only
 *          if Temporary_index(item) is in [0, maxIndex] and
 *          sparse[Temporary_index(item)] == item.
 *
 * root_: This array contains for each entry a pointer to the equivalence
 *        class this element belongs to.
 *
 * maxIndex: The maximum index over all numbered keys.
 *
 * maxSize: The maximum number of numbered Temporary this structure
 *          can handle. When the limit is reached, the structure will
 *          be automatically resized to handle more Temporaries.
 */

#line 35 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
#include "GRA/GRA_.h"

#line 40 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 58 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static int
Equiv_Item_index(Equiv_Item item)
{
  return Temporary_index(item);
}

#line 64 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
typedef struct Equiv_Tree_ {
  IDList_ root_;	// List of Equiv_Item in the set
  Equiv_Item rep;	// Representative for the set
} Equiv_Tree_;

#line 69 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
typedef struct Equiv_ {
  Memory memory;
  int initialSize;
  int maxIndex;
  int maxSize;
  IArray_ sparse_;
  IArray_ root_;
  Equiv_Item *sparse;
  Equiv_Tree *root;
} Equiv_;

#line 80 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
void
Equiv_ctor(Equiv equiv, int initialSize, Memory memory) {
  equiv->memory = memory;
  equiv->initialSize = initialSize;
  equiv->maxIndex = -1;
  equiv->maxSize = equiv->initialSize == 0? 64: equiv->initialSize;
  IArray_Ctor((IArray)&equiv->sparse_, equiv->memory, equiv->maxSize, sizeof(Equiv_Item));
  IArray_Ctor((IArray)&equiv->root_, equiv->memory, equiv->maxSize, sizeof(Equiv_Tree));
  equiv->sparse = (Equiv_Item *)IArray_base((IArray)&equiv->sparse_);
  equiv->root = (Equiv_Tree *)IArray_base((IArray)&equiv->root_);
}

#line 92 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static void
Equiv_resize(Equiv equiv, int min_size) {
  int step = equiv->initialSize == 0? 64: equiv->initialSize;
  equiv->maxSize += step;
  if (equiv->maxSize < min_size + step) {
    equiv->maxSize = min_size + step;
  }
  IArray_resize((IArray)&equiv->sparse_, equiv->maxSize);
  IArray_resize((IArray)&equiv->root_, equiv->maxSize);
  equiv->sparse = (Equiv_Item *)IArray_base((IArray)&equiv->sparse_);
  equiv->root = (Equiv_Tree *)IArray_base((IArray)&equiv->root_);
}

#line 105 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
Equiv
Equiv_new(int initialSize, Memory memory) {
  Equiv equiv;
  equiv = Memory_alloc(memory, sizeof(Equiv_));
  Equiv_ctor(equiv, initialSize, memory);
  return equiv;
}

#line 113 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
void
Equiv_dtor(Equiv equiv) {
  IArray_Dtor((IArray)&equiv->sparse_);
  IArray_Dtor((IArray)&equiv->root_);
}

#line 119 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
void
Equiv_del(Equiv equiv) {
  Equiv_dtor(equiv);
  Memory_free(equiv->memory, equiv);
}

#line 125 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static bool
Equiv_isValidItem(Equiv equiv, Equiv_Item item)
{
  return Equiv_Item_index(item) >= 0 && 
    Equiv_Item_index(item) <= equiv->maxIndex &&
    equiv->sparse[Equiv_Item_index(item)] == item;
}

#line 134 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
/* Check if key already belongs to a Equiv. Returns the address of the
   cell that holds the Equiv this key belongs to, or NULL otherwise.
   Note that this function has no side effect. The returned address
   may no longer be valid after a Equiv lookup which may resize the
   Equiv. */

#line 140 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static Equiv_Tree *
Equiv_search(Equiv equiv, Equiv_Item key) {
  if (Equiv_isValidItem(equiv, key))
    return equiv->root+Equiv_Item_index(key);
  return NULL;
}

#line 147 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
/* Check if key already belongs to a Equiv. Returns the address of the
   cell that hold the Equiv this key belongs to. If key was not already
   in a Equiv, a new entry is created for this key, with the Equiv it
   belongs to being NULL. The returned address may no longer be valid
   after a new Equiv lookup which may resize the Equiv. */

#line 153 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static Equiv_Tree *
Equiv_lookup(Equiv equiv, Equiv_Item key) {
  if (Equiv_isValidItem(equiv, key))
    return equiv->root+Equiv_Item_index(key);
  
  /* Insert key */
  // Assume key is available: Equiv_Item_setIndex(key, equiv->maxIndex++);
  if (Equiv_Item_index(key) >= equiv->maxSize)
    Equiv_resize(equiv, Equiv_Item_index(key)+1);
  if (Equiv_Item_index(key) > equiv->maxIndex)
    equiv->maxIndex = Equiv_Item_index(key);
  equiv->sparse[Equiv_Item_index(key)] = key;
  return equiv->root+Equiv_Item_index(key);
}

#line 168 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
/*
 * Equiv_Tree
 *
 * Defines the set of node for an equivalence tree.
 */
static void
Equiv_Tree_ctor(Equiv_Tree tree, Memory memory)
{
  IDList_Ctor((IDList)&tree->root_, memory);
  tree->rep = NULL;
}

#line 180 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static Equiv_Tree
Equiv_Tree_new(Memory memory) {
  Equiv_Tree tree = Memory_alloc(memory, sizeof(Equiv_Tree_));
  Equiv_Tree_ctor(tree, memory);
  return tree;
}

#line 187 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static void
Equiv_Tree_dtor(Equiv_Tree tree) {
  IDList_Dtor((IDList)&tree->root_);
  tree->rep = NULL;
}

#line 193 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
static void
Equiv_Tree_del(Equiv_Tree tree) {
  Equiv_Tree_dtor(tree);
}

#line 198 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
#define Equiv_Tree_setRep(tree, item) (tree)->rep = item
#define Equiv_Tree_getRep(tree) (tree)->rep
#define Equiv_Tree_insert(tree, item) IDList_PUSH((IDList)&(tree)->root_, Equiv_Item, item)
#define Equiv_Tree_count(tree) IDList_count((IDList)&(tree)->root_)
#define Equiv_Tree_clear(tree) IDList_empty((IDList)&(tree)->root_, NULL)
#define Equiv_Tree_FOREACH_Item(tree, item) IDList_FOREACH((IDList)&(tree)->root_, Equiv_Item, ITEM_PTR) { Equiv_Item item = *ITEM_PTR;
#define Equiv_Tree_ENDEACH_Item } IDList_ENDEACH

#line 206 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
Equiv_Tree
Equiv_make(Equiv equiv, Equiv_Item key) {
  Equiv_Tree *root = Equiv_lookup(equiv, key);
  Except_CHECK(*root == NULL);
  *root = Equiv_Tree_new(equiv->memory);
  Equiv_Tree_insert(*root, key);
  Equiv_Tree_setRep(*root, key);
  return *root;
}

#line 216 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
Equiv_Item
Equiv_find(Equiv equiv, Equiv_Item key) {
  Equiv_Tree *root = Equiv_search(equiv, key);
  return root? Equiv_Tree_getRep(*root): key;
}

#line 222 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
Equiv_Item
Equiv_union(Equiv equiv, Equiv_Item keyX, Equiv_Item keyY) {
  Equiv_Tree *rootX, *rootY;
  rootX = Equiv_lookup(equiv, keyX);
  rootY = Equiv_lookup(equiv, keyY);
  // rootX may no longer be valid if equiv have been reallocated between
  // the two lookup. Get now a valid pointer.
  rootX = Equiv_search(equiv, keyX);

#line 231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  if (*rootX && *rootY) {
    Equiv_Tree root;
    if (Equiv_Tree_count(*rootX) < Equiv_Tree_count(*rootY)) {
      Equiv_Tree *tmp = rootY;
      rootY = rootX;
      rootX = tmp;
    }
    root = *rootY;
    Equiv_Tree_FOREACH_Item(root, item) {
      Equiv_Tree_insert(*rootX, item);
      *Equiv_lookup(equiv, item) = *rootX;
    } Equiv_Tree_ENDEACH_Item;
    Equiv_Tree_del(root);
  }

#line 246 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  else if (*rootX && !*rootY) {
    Equiv_Tree_insert(*rootX, keyY);
    *rootY = *rootX;
  }

#line 251 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  else if (!*rootX && *rootY) {
    Equiv_Tree_insert(*rootY, keyX);
    *rootX = *rootY;
  }

#line 256 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  else /* !rootX && !rootY */ {
    *rootX = *rootY = Equiv_make(equiv, keyX);
    Equiv_Tree_insert(*rootX, keyY);
  }

#line 261 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  Except_CHECK(*rootX == *rootY);
  Except_CHECK(Equiv_find(equiv, keyX) == Equiv_find(equiv, keyY));
  Equiv_Tree_FOREACH_Item(*rootX, item) {
    Except_CHECK(Equiv_find(equiv, keyX) == Equiv_find(equiv, item));
  } Equiv_Tree_ENDEACH_Item;

#line 267 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
  return Equiv_Tree_getRep(*rootX);
}

#line 270 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 282 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
#if XCC__C
static void
Equiv_TEST(void)
{
#include "GRA/Equiv_.c"
}

#line 289 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Equiv.xcc"
int
main(int argc, char **argv)
{
  GRA_Test(Equiv_TEST);
  return 0;
}
#endif


#endif//Equiv_c_INCLUDED
