#ifndef InterfGraph__h_INCLUDED
#define InterfGraph__h_INCLUDED

//
#include "GRA/InterfGraph.h"

//
static inline uint32_t
BitMatrix_size(uint32_t n)
{
  return (n * (n+1))/2;
}

//
static inline uint32_t
BitMatrix_index(uint32_t i, uint32_t j)
{
  // The bit matrix is triangular, we have (i,j) included iff (j,i) is included.
  if (i < j)
    return i + (j*(j+1))/2;
  else
    return j + (i*(i+1))/2;
}

//
/**
 * To know in which list a node is.
 */
typedef enum {
  noneNode,
  _precolored,
  _initial,
  _simplifyWorklist,
  _freezeWorklist,
  _spillWorklist,
  _spilledNodes,
  _coalescedNodes,
  _coloredNodes,
  _selectStack,
  _discardedNodes,
} NodeBelongTo;

//
/**
 * To know in which list a move is.
 */
typedef enum {
  noneMove,
  _coalescedMoves,
  _constrainedMoves,
  _frozenMoves,
  _worklistMoves,
  _activeMoves,
} MoveBelongTo;

//
/**
 * The structure of a move in the movelist of a node.
 */
struct InterfMove_ {
  //@args Operation operation, struct InterfNode_ *def,
  //@args struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost
  Operation OPERATION; //!< Related operation.
  struct InterfNode_ *DEF; //!< Related def
  struct InterfNode_ *USE; //!< Related use
  float COST; //!< Related cost
  MoveBelongTo BELONGTO; //!< In which set it is.
  int32_t HIGH; //!< Lower bound on the number of high-degree neighbors (see MLRISC)
  int32_t DEFIDX;
  int32_t USEIDX;
};

extern InterfMove
InterfMove_Ctor(InterfMove this, Operation operation, struct InterfNode_ *def, struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost);

extern InterfMove
InterfMove_Copy(InterfMove this, const_InterfMove that);

extern void
InterfMove_Dtor(InterfMove this);

extern size_t
InterfMove_Size(Operation operation, struct InterfNode_ *def, struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost);

#define InterfMove_OPERATION(this) ((this)->OPERATION)
#define InterfMove__OPERATION(this) (&(this)->OPERATION)
#define InterfMove_DEF(this) ((this)->DEF)
#define InterfMove__DEF(this) (&(this)->DEF)
#define InterfMove_USE(this) ((this)->USE)
#define InterfMove__USE(this) (&(this)->USE)
#define InterfMove_COST(this) ((this)->COST)
#define InterfMove__COST(this) (&(this)->COST)
#define InterfMove_BELONGTO(this) ((this)->BELONGTO)
#define InterfMove__BELONGTO(this) (&(this)->BELONGTO)
#define InterfMove_HIGH(this) ((this)->HIGH)
#define InterfMove__HIGH(this) (&(this)->HIGH)
#define InterfMove_DEFIDX(this) ((this)->DEFIDX)
#define InterfMove__DEFIDX(this) (&(this)->DEFIDX)
#define InterfMove_USEIDX(this) ((this)->USEIDX)
#define InterfMove__USEIDX(this) (&(this)->USEIDX)


//
/**
 * Print (human readable) a move.
 * \param file The file descriptor.
 * \param interfMove The move.
 */
bool
InterfMove_pretty(InterfMove interfMove, FILE* file);

//
/**
 * Structure of a node, in the interference graph, corresponding to a temporary and other informations.
 */
struct InterfNode_ {
  //@args Memory memory, Temporary temporary
  Memory MEMORY; //!< Memory for allocation
  Temporary TEMPORARY; //!< Related temporary.
  IDList ADJLIST; //!< Set of nodes that interfere with this one.
  struct InterfNode_ *ALIAS; //!< Node coalesced with.
  NodeBelongTo BELONGTO; //!< In which set it is.
  IDList MOVELIST; //!< List of moves associated with this node.
  int32_t DEGREE; //!< Degree of this node.
  Register COLOR; //!< Color assigned (nonfinal).
  IDList USES;
  IDList DEFS;
  float SPILLCOST;
  bool FLAGS;
};

extern InterfNode
InterfNode_Ctor(InterfNode this, Memory memory, Temporary temporary);

extern InterfNode
InterfNode_Copy(InterfNode this, const_InterfNode that);

extern void
InterfNode_Dtor(InterfNode this);

extern size_t
InterfNode_Size(Memory memory, Temporary temporary);

#define InterfNode_MEMORY(this) ((this)->MEMORY)
#define InterfNode__MEMORY(this) (&(this)->MEMORY)
#define InterfNode_TEMPORARY(this) ((this)->TEMPORARY)
#define InterfNode__TEMPORARY(this) (&(this)->TEMPORARY)
#define InterfNode_ADJLIST(this) ((this)->ADJLIST)
#define InterfNode__ADJLIST(this) (&(this)->ADJLIST)
#define InterfNode_ALIAS(this) ((this)->ALIAS)
#define InterfNode__ALIAS(this) (&(this)->ALIAS)
#define InterfNode_BELONGTO(this) ((this)->BELONGTO)
#define InterfNode__BELONGTO(this) (&(this)->BELONGTO)
#define InterfNode_MOVELIST(this) ((this)->MOVELIST)
#define InterfNode__MOVELIST(this) (&(this)->MOVELIST)
#define InterfNode_DEGREE(this) ((this)->DEGREE)
#define InterfNode__DEGREE(this) (&(this)->DEGREE)
#define InterfNode_COLOR(this) ((this)->COLOR)
#define InterfNode__COLOR(this) (&(this)->COLOR)
#define InterfNode_USES(this) ((this)->USES)
#define InterfNode__USES(this) (&(this)->USES)
#define InterfNode_DEFS(this) ((this)->DEFS)
#define InterfNode__DEFS(this) (&(this)->DEFS)
#define InterfNode_SPILLCOST(this) ((this)->SPILLCOST)
#define InterfNode__SPILLCOST(this) (&(this)->SPILLCOST)
#define InterfNode_FLAGS(this) ((this)->FLAGS)
#define InterfNode__FLAGS(this) (&(this)->FLAGS)


//
/**
 * InterfNode flags
 */
#define InterfNode_POTENTIALSPILL 1
#define InterfNode_ACTUALSPILL 2
#define InterfNode_ALREADYSPILLED 4

//
/**
 * Set a flag mask for the node.
 */
static inline void
InterfNode_setFlag(InterfNode interfNode, int flags);
//
static inline void
InterfNode_setFlag(InterfNode interfNode, int flags)
{
  *InterfNode__FLAGS(interfNode) |= flags;
}

//
/**
 * Reset a flag mask for the node.
 */
static inline void
InterfNode_resetFlag(InterfNode interfNode, int flags);
//
static inline void
InterfNode_resetFlag(InterfNode interfNode, int flags)
{
  *InterfNode__FLAGS(interfNode) &= ~flags;
}

//
/**
 * Get interfnode flags.
 */
static inline int
InterfNode_flags(InterfNode interfNode);
//
static inline int
InterfNode_flags(InterfNode interfNode)
{
  return InterfNode_FLAGS(interfNode);
}

//
/**
 * Test interfnode flag mask.
 */
static inline bool
InterfNode_hasFlag(InterfNode interfNode, int flag);
//
static inline bool
InterfNode_hasFlag(InterfNode interfNode, int flag)
{
  return InterfNode_FLAGS(interfNode) & flag;
}

//
/**
 * Set the node coalesced with.
 * \param interfNode1 The interference node.
 * \param interfNode2 The interference node coalesced with.
 */
static inline void
InterfNode_setAlias(InterfNode interfNode1, InterfNode interfNode2);
//
static inline void
InterfNode_setAlias(InterfNode interfNode1, InterfNode interfNode2)
{
	*InterfNode__ALIAS(interfNode1) = interfNode2;
}

//
/**
 * Get the node coalesced with.
 * \param interfNode The interference node.
 * \return The interference node coalesced with.
 */
static inline InterfNode
InterfNode_getAlias(InterfNode interfNode);
//
static inline InterfNode
InterfNode_getAlias(InterfNode interfNode)
{
	if( InterfNode_BELONGTO(interfNode) == _coalescedNodes )
		return InterfNode_getAlias(InterfNode_ALIAS(interfNode));
	else return interfNode;
}

//
/**
 * To know if a node is related to a precolored temporary.
 * \param interfNode The interference node.
 * \return \e true if the node is related to a precolored temporary, else \e false.
 */
static inline bool
InterfNode_isPrecolored(InterfNode interfNode);
//
static inline bool
InterfNode_isPrecolored(InterfNode interfNode)
{
	return Temporary_isDedicated(InterfNode_TEMPORARY(interfNode));
}

//
/**
 * Assign a physical register to the node.
 * \param interfNode The interference node.
 * \param color The physical register.
 */
static inline void
InterfNode_setColor(InterfNode interfNode, Register color);
//
static inline void
InterfNode_setColor(InterfNode interfNode, Register color)
{
	*InterfNode__COLOR(interfNode) = color;
}

//
static inline uint32_t
InterfGraph_temporaryIdentity(Temporary tempo) {
/*   RegFile regFile = Temporary_regFile(tempo); */
/*   uint32_t result = Temporary_identity(tempo); */
/*   if(regFile != 0 && Temporary_isDedicated(tempo)) { */
/*     uint32_t offset = RegFile_highRegister(regFile - 1); */
/*     result = Temporary_register(tempo) - offset - 1; */
/*   } */
  return Temporary_index(tempo);
}

//
/**
 * Increment the degree of a node.
 * \param interfNode The interference node.
 * \return The incremented degree.
 */
static inline uint32_t
InterfNode_incDegree(InterfNode interfNode);
//
static inline uint32_t
InterfNode_incDegree(InterfNode interfNode)
{
	*InterfNode__DEGREE(interfNode) = InterfNode_DEGREE(interfNode)+1;
	return InterfNode_DEGREE(interfNode);
}

//
/**
 * Add a node to the adjacency list.
 * \param interfNode The first interference node.
 * \param interfWith The interference node to be added into the adjacency list of the first one.
 */
static inline void
InterfNode_addToAdjList( InterfNode interfNode,
			 InterfNode interfWith);
//
static inline void
InterfNode_addToAdjList( InterfNode interfNode,
			 InterfNode interfWith)
{
	IDList_PUSH( InterfNode_ADJLIST(interfNode), InterfNode, interfWith);
}

//
static inline void
InterfNode_addToMoveList( InterfNode interfNode,
			  InterfMove interfMove);
//
static inline void
InterfNode_addToMoveList( InterfNode interfNode,
			  InterfMove interfMove)
{
	IDList_PUSH( InterfNode_MOVELIST(interfNode), InterfMove, interfMove);
}

//
/**
 * Print (human readable) a node.
 * \param file The file descriptor.
 * \param interfNode The interference node.
 */
bool
InterfNode_pretty(InterfNode interfNode, FILE *file);

/*
 * Create the adjacency matrix of the graph.
 * \param interfGraph The interference graph.
 * \param count The maximal number of nodes.
 */
void
InterfGraph_createAdjSet(InterfGraph interfGraph, uint32_t count);

//
/**
 * Reallocate the adjacency matrix for the graph.
 * The IntrfNodes list is preserved and enlarged.
 * \param interfGraph The interference graph.
 * \param count The maximal number of nodes, must be greater than the current count.
 */
static inline void
InterfGraph_reallocAdjSet(InterfGraph interfGraph, uint32_t count);
//
static inline void
InterfGraph_reallocAdjSet(InterfGraph interfGraph, uint32_t count)
{
  int old_size = InterfGraph_SIZE(interfGraph);
  ASSERT(count >= old_size);
  if (count > old_size) {
    *InterfGraph__SIZE(interfGraph) = count;
    BitSet_Dtor(InterfGraph_ADJSET(interfGraph));
    BitSet_Ctor(InterfGraph_ADJSET(interfGraph), InterfGraph_MEMORY(interfGraph),
                BitMatrix_size(count));
    *InterfGraph__NODES(interfGraph) = Memory_realloc(InterfGraph_MEMORY(interfGraph), InterfGraph_NODES(interfGraph), count*sizeof(InterfNode));
    memset(&InterfGraph_NODES(interfGraph)[old_size], 0, (count-old_size)*sizeof(InterfNode));
  } else {
    BitSet_empty(InterfGraph_ADJSET(interfGraph));
  }
}

//
/**
 * To know if two node indexes are neighbors in the interference graph.
 * \param interfGraph The interference graph.
 * \param i The first interference node index.
 * \param j The second interference node index.
 * \return \e true if these two nodes are neighbors, else \e false.
 */
static inline bool
InterfGraph_isAdjacentIndex(InterfGraph interfGraph,
                            uint32_t i,
                            uint32_t j);
//
static inline bool
InterfGraph_isAdjacentIndex(InterfGraph interfGraph,
			    uint32_t i, uint32_t j)
{
	return BitSet_contains(InterfGraph_ADJSET(interfGraph),
			       BitMatrix_index(i, j));
}

//
/**
 * Add an InterfNode to the interference graph.
 * \param interfGraph The interference graph.
 * \param interfNode The node to add to the graph.
 */
static inline void
InterfGraph_addNode(InterfGraph interfGraph, InterfNode interfNode);
//
static inline void
InterfGraph_addNode(InterfGraph interfGraph, InterfNode interfNode)
{
	InterfGraph_NODES(interfGraph)[InterfNode_index(interfNode)] = interfNode;
}

//
/**
 * Get a node from its index.
 * \param interfGrapg The interference graph.
 * \param index the node index
 * \return The node correspondif to the inex
 * To get the node from a temporary do
 * InterfGraph_addNode(graph, Temporary_identity(temporary)
 */
static inline InterfNode
InterfGraph_getNode(InterfGraph interfGraph, int index);
//
//FIXME: For dedicated, we should use on representative by register
static inline InterfNode
InterfGraph_getNode(InterfGraph interfGraph, int index)
{
	return InterfGraph_NODES(interfGraph)[index];
}

//
/**
 * Add an interference edge in the undirected interference graph.
 * The implementation uses a triangular bit matrix.
 * \param interfGraph The interference graph.
 * \param interfNode1 The first interference node.
 * \param interfNode2 The second interference node.
 * \return Always \e true.
 */
static inline void
InterfGraph_addToAdjSet(InterfGraph interfGraph,
		    InterfNode interfNode1,
		    InterfNode interfNode2 );
//
static inline void
InterfGraph_addToAdjSet(InterfGraph interfGraph,
		    InterfNode interfNode1,
		    InterfNode interfNode2)
{
	BitSet_insert(InterfGraph_ADJSET(interfGraph),
		      BitMatrix_index(InterfNode_index(interfNode1), InterfNode_index(interfNode2)));
}

//
/**
 * Print (human readable) a graph.
 * \param file The file descriptor.
 * \param interfGraph The interference graph.
 */
bool
InterfGraph_pretty(InterfGraph interfGraph, FILE *file);

//
struct SplitInfo_ {
  //@args Operation point, Temporary splitVar, Temporary origVar, int32_t opndIdx
  Operation POINT;
  Temporary SPLITVAR;
  Temporary ORIGVAR;
  int32_t OPERANDIDX;
};
typedef struct SplitInfo_ SplitInfo_, *SplitInfo;
typedef const struct SplitInfo_ *const_SplitInfo;
typedef struct SplitInfo_ * restrict_SplitInfo;

extern SplitInfo
SplitInfo_Ctor(SplitInfo this, Operation point, Temporary splitVar, Temporary origVar, int32_t opndIdx);

extern SplitInfo
SplitInfo_Copy(SplitInfo this, const_SplitInfo that);

extern void
SplitInfo_Dtor(SplitInfo this);

extern size_t
SplitInfo_Size(Operation point, Temporary splitVar, Temporary origVar, int32_t opndIdx);

#define SplitInfo_POINT(this) ((this)->POINT)
#define SplitInfo__POINT(this) (&(this)->POINT)
#define SplitInfo_SPLITVAR(this) ((this)->SPLITVAR)
#define SplitInfo__SPLITVAR(this) (&(this)->SPLITVAR)
#define SplitInfo_ORIGVAR(this) ((this)->ORIGVAR)
#define SplitInfo__ORIGVAR(this) (&(this)->ORIGVAR)
#define SplitInfo_OPERANDIDX(this) ((this)->OPERANDIDX)
#define SplitInfo__OPERANDIDX(this) (&(this)->OPERANDIDX)


/*
 * InterfGraph_INIT --	Initialize the InterfGraph module.
 */
#define InterfGraph_INIT()

/*
 * InterfGraph_FINI --	Finalize the InterfGraph module.
 */
#define InterfGraph_FINI()


#endif//InterfGraph__h_INCLUDED
