#ifndef FastList__h_INCLUDED
#define FastList__h_INCLUDED

//
#include "SCD/FastList.h"

//
struct FastListRank_ {
  //@args	int release
  uint32_t CAPACITY;		// The resource capacity of this FastListRank.
  uint32_t CONSUMED;		// The resource consumed in this FastListRank.
  int16_t RELEASE;		// The release time of this FastListRank.
  FastListRank PARENT;		// The parent of this FastListRank for the Union-Find.
  //@access NEXT	((FastListRank)(this) + 1)
};

extern FastListRank
FastListRank_Ctor(FastListRank this, int release);

extern FastListRank
FastListRank_Copy(FastListRank this, const_FastListRank that);

extern void
FastListRank_Dtor(FastListRank this);

extern size_t
FastListRank_Size(int release);

#define FastListRank_CAPACITY(this) ((this)->CAPACITY)
#define FastListRank__CAPACITY(this) (&(this)->CAPACITY)
#define FastListRank_CONSUMED(this) ((this)->CONSUMED)
#define FastListRank__CONSUMED(this) (&(this)->CONSUMED)
#define FastListRank_RELEASE(this) ((this)->RELEASE)
#define FastListRank__RELEASE(this) (&(this)->RELEASE)
#define FastListRank_PARENT(this) ((this)->PARENT)
#define FastListRank__PARENT(this) (&(this)->PARENT)
#define FastListRank_NEXT(this) (((FastListRank)(this) + 1))


/*
 * FastListRank_find --	Find of the Union-Find data structure.
 */
static inline FastListRank
FastListRank_find(FastListRank this)
{
  FastListRank root = this, that = this;
  while (FastListRank_PARENT(root) != NULL) {
    root = FastListRank_PARENT(root);
  }
  while (that != root) {
    FastListRank parent = FastListRank_PARENT(that);
    *FastListRank__PARENT(that) = root;
    that = parent;
  }
  return root;
}

/*
 * FastListRank_schedule --	Schedule at this FastListRank.
 */
unsigned
FastListRank_schedule(FastListRank this, unsigned available, unsigned required);

/*
 * FastListRank_scheduleS --	Specialize FastListRank_schedule when available = 2^shift.
 */
unsigned
FastListRank_scheduleS(FastListRank this, int shift, unsigned required);

//
struct FastListTask_ {
  //@args	const uint8_t *requirements, RelaxationItem item
  const uint8_t *REQUIREMENTS;		// The resource requirements of this FastListTask.
  struct RelaxationItem_ *ITEM;		// The corresponding RelaxationItem.
  //@access INDEX	RelaxationItem_INDEX(FastListTask_ITEM(this))
  FastListRank RANK;			// The FastListRank of this FastListTask.
  FastListTask LINKS[FastListType__];	// The links to order FastListTask(s) by increasing date.
  int16_t DATES[FastListType__];	// The release date and the deadline date.
  //@access RELEASE	(FastListTask_DATES(this)[FastListType_Release])
  //@mutate RELEASE	(FastListTask_DATES(this) + FastListType_Release)
  //@access SCHEDULE	(FastListTask_DATES(this)[FastListType_Schedule])
  //@mutate SCHEDULE	(FastListTask_DATES(this) + FastListType_Schedule)
  //@access DEADLINE	(FastListTask_DATES(this)[FastListType_Deadline])
  //@mutate DEADLINE	(FastListTask_DATES(this) + FastListType_Deadline)
};

extern FastListTask
FastListTask_Ctor(FastListTask this, const uint8_t *requirements, RelaxationItem item);

extern FastListTask
FastListTask_Copy(FastListTask this, const_FastListTask that);

extern void
FastListTask_Dtor(FastListTask this);

extern size_t
FastListTask_Size(const uint8_t *requirements, RelaxationItem item);

#define FastListTask_REQUIREMENTS(this) ((this)->REQUIREMENTS)
#define FastListTask__REQUIREMENTS(this) (&(this)->REQUIREMENTS)
#define FastListTask_ITEM(this) ((this)->ITEM)
#define FastListTask__ITEM(this) (&(this)->ITEM)
#define FastListTask_RANK(this) ((this)->RANK)
#define FastListTask__RANK(this) (&(this)->RANK)
#define FastListTask_LINKS(this) ((this)->LINKS)
#define FastListTask__LINKS(this) (&(this)->LINKS)
#define FastListTask_DATES(this) ((this)->DATES)
#define FastListTask__DATES(this) (&(this)->DATES)
#define FastListTask_INDEX(this) (RelaxationItem_INDEX(FastListTask_ITEM(this)))
#define FastListTask_DEADLINE(this) ((FastListTask_DATES(this)[FastListType_Deadline]))
#define FastListTask_RELEASE(this) ((FastListTask_DATES(this)[FastListType_Release]))
#define FastListTask_SCHEDULE(this) ((FastListTask_DATES(this)[FastListType_Schedule]))
#define FastListTask__DEADLINE(this) ((FastListTask_DATES(this) + FastListType_Deadline))
#define FastListTask__RELEASE(this) ((FastListTask_DATES(this) + FastListType_Release))
#define FastListTask__SCHEDULE(this) ((FastListTask_DATES(this) + FastListType_Schedule))


/*
 * FastListTask_link --	Used by FastList_FOREACH_(Type)_FastListTask.
 */
static inline FastListTask
FastListTask_link(FastListTask this, FastListType type)
{
  return FastListTask_LINKS(this)[type];
}

/*
 * FastListTask_compare --	Compare two FastListTask(s) on their DATES.
 */
static inline int
FastListTask_compare(FastListTask this, FastListTask that, FastListType type)
{
  return FastListTask_DATES(this)[type] - FastListTask_DATES(that)[type];
}

/*
 * FastListTask_meld --	Meld function for the merge-sort of FastListTasks.
 */
FastListTask
FastListTask_meld(FastListTask this, FastListTask that, FastListType type);

//
struct FastList_ {
  //@args	Memory memory, int32_t count, const uint8_t *availability
  const uint8_t *AVAILABILITY;	// The resource availability.
  FastListTask_ *TASKS;		// Array of FastListTasks in creation order.
  //@access FIRSTTASK	FastList_TASKS(this)
  FastListTask ENDTASK;		// The end FastListTask in this TASKS.
  //@access COUNT	(FastList_ENDTASK(this) - FastList_TASKS(this))
  FastListRank_ *RANKS;		// The FastListRanks in increasing RELEASE time.
  FastListRank ENDRANK;		// The end FastListRank in this RANKS.
  FastListTask_ LISTS_;			// The list of FastListTask(s) by increasing date.
  //@access LISTS	FastList__LISTS_(this)
  //@access LINKS	FastListTask_LINKS(FastList_LISTS(this))
  IQueue_ IQUEUE_;		// The IQueue used for sorting the LISTS.
  //@access IQUEUE	FastList__IQUEUE_(this)
  //@access MEMORY	IQueue_memory(FastList_IQUEUE(this))
};

extern FastList
FastList_Ctor(FastList this, Memory memory, int32_t count, const uint8_t *availability);

extern FastList
FastList_Copy(FastList this, const_FastList that);

extern void
FastList_Dtor(FastList this);

extern size_t
FastList_Size(Memory memory, int32_t count, const uint8_t *availability);

#define FastList_AVAILABILITY(this) ((this)->AVAILABILITY)
#define FastList__AVAILABILITY(this) (&(this)->AVAILABILITY)
#define FastList_TASKS(this) ((this)->TASKS)
#define FastList__TASKS(this) (&(this)->TASKS)
#define FastList_ENDTASK(this) ((this)->ENDTASK)
#define FastList__ENDTASK(this) (&(this)->ENDTASK)
#define FastList_RANKS(this) ((this)->RANKS)
#define FastList__RANKS(this) (&(this)->RANKS)
#define FastList_ENDRANK(this) ((this)->ENDRANK)
#define FastList__ENDRANK(this) (&(this)->ENDRANK)
#define FastList_LISTS_(this) ((this)->LISTS_)
#define FastList__LISTS_(this) (&(this)->LISTS_)
#define FastList_IQUEUE_(this) ((this)->IQUEUE_)
#define FastList__IQUEUE_(this) (&(this)->IQUEUE_)
#define FastList_FIRSTTASK(this) (FastList_TASKS(this))
#define FastList_IQUEUE(this) (FastList__IQUEUE_(this))
#define FastList_LINKS(this) (FastListTask_LINKS(FastList_LISTS(this)))
#define FastList_MEMORY(this) (IQueue_memory(FastList_IQUEUE(this)))
#define FastList_LISTS(this) (FastList__LISTS_(this))
#define FastList_COUNT(this) ((FastList_ENDTASK(this) - FastList_TASKS(this)))


/*
 * FastList_tasks --	Used by FastList_FOREACH_FastListTask.
 */
static inline FastListTask_ *
FastList_tasks(FastList this)
{
  return FastList_TASKS(this);
}

/*
 * FastList_endTask --	Used by FastList_FOREACH_FastListTask.
 */
static inline FastListTask
FastList_endTask(FastList this)
{
  return FastList_ENDTASK(this);
}

/*
 * FastList_FOREACH_FastListTask --	Iterate over this FastList TASKS in index order.
 */
#define FastList_FOREACH_FastListTask(this, task) { \
  FastListTask task = FastList_tasks(this); \
  FastListTask FastList_ENDTASK = FastList_endTask(this); \
  for (; task < FastList_ENDTASK; task++) {
#define FastList_ENDEACH_FastListTask \
  } \
}

/*
 * FastList_ranks --	Used by FastList_FOREACH_FastListRank.
 */
static inline FastListRank_ *
FastList_ranks(FastList this)
{
  return FastList_RANKS(this);
}

/*
 * FastList_endRank --	Used by FastList_FOREACH_FastListRank.
 */
static inline FastListRank
FastList_endRank(FastList this)
{
  return FastList_ENDRANK(this);
}

/*
 * FastList_FOREACH_FastListRank --	Iterate over this FastList RANKS in RELEASE order.
 */
#define FastList_FOREACH_FastListRank(this, rank) { \
  FastListRank rank = FastList_ranks(this); \
  FastListRank FastList_ENDRANK = FastList_endRank(this); \
  for (; rank < FastList_ENDRANK; rank++) {
#define FastList_ENDEACH_FastListRank \
  } \
}

/*
 * FastList_list --	Used by FastList_FOREACH_(Type)_FastListTask.
 */
static inline FastListTask
FastList_list(FastList this, FastListType type)
{
  return FastList_LINKS(this)[type];
}

/*
 * FastList_FOREACH_Release_FastListTask --	Iterate over this FastList TASKS in Release order.
 */
#define FastList_FOREACH_Release_FastListTask(this, task) { \
  FastListTask task = FastList_list(this, FastListType_Release); \
  for (; task != NULL; task = FastListTask_link(task, FastListType_Release)) {
#define FastList_ENDEACH_Release_FastListTask \
  } \
}

/*
 * FastList_FOREACH_Deadline_FastListTask --	Iterate over this FastList TASKS in Deadline order.
 */
#define FastList_FOREACH_Deadline_FastListTask(this, task) { \
  FastListTask task = FastList_list(this, FastListType_Deadline); \
  for (; task != NULL; task = FastListTask_link(task, FastListType_Deadline)) {
#define FastList_ENDEACH_Deadline_FastListTask \
  } \
}

/*
 * FastList_FOREACH_Schedule_FastListTask --	Iterate over this FastList TASKS in Schedule order.
 */
#define FastList_FOREACH_Schedule_FastListTask(this, task) { \
  FastListTask task = FastList_list(this, FastListType_Schedule); \
  for (; task != NULL; task = FastListTask_link(task, FastListType_Schedule)) {
#define FastList_ENDEACH_Schedule_FastListTask \
  } \
}

/*
 * FastList_makeTask --	Create a new FastListTask in this TASKS array.
 */
static inline FastListTask
FastList_makeTask(FastList this, const uint8_t *requirements, RelaxationItem item)
{
  FastListTask task = FastList_ENDTASK(this);
  int index = task - FastList_TASKS(this);
  FastListTask_Ctor(task, requirements, item);
  if (index > 0) {
    // Not the first FastListTask created.
    FastListTask prev_task = FastList_TASKS(this) + index - 1;
    FastListTask_LINKS(prev_task)[FastListType_Release] = task;
    FastListTask_LINKS(prev_task)[FastListType_Schedule] = task;
    FastListTask_LINKS(prev_task)[FastListType_Deadline] = task;
  } else {
    // This is the first FastListTask created.
    FastList_LINKS(this)[FastListType_Release] = task;
    FastList_LINKS(this)[FastListType_Schedule] = task;
    FastList_LINKS(this)[FastListType_Deadline] = task;
  }
  ++*FastList__ENDTASK(this);
  return task;
}

/*
 * FastList_makeRank --	Create a new FastListRank in this RANKS array.
 */
FastListRank
FastList_makeRank(FastList this, int release);

/*
 * FastList_makeRanks --	Make this FastList RANKS.
 */
void
FastList_makeRanks(FastList this, unsigned available);

/*
 * FastList_INIT --	Initialize the FastList module.
 */
#define FastList_INIT()

/*
 * FastList_FINI --	Finalize the FastList module.
 */
#define FastList_FINI()


#endif//FastList__h_INCLUDED
