#ifndef DDGraph__h_INCLUDED
#define DDGraph__h_INCLUDED

//
#include "XFA/DDGraph.h"

/*
 * DDGraph_makeLifeNode --	Return the lifetime node associated to =prod_node=.
 *
 * If the lifetime node does not exist, it is flow requirements are initialized,
 * and the flow requirements of the producer node are updated.
 */
DDGraphNode
DDGraph_makeLifeNode(DDGraph this, DDGraphNode prod_node, Temporary temporary);

/*
 * DDGraph_addRegisterRAW --	Add a RAW register dependence arc between two DDGraphNodes.
 *
 * If =lifeOK=, a =life_node= is created and a Life arc is added between
 * =head_node= and =life_node=. This has the effect of pulling =tail_node= and
 * =head_node= together, when the network flow problem is solved, see B. Dupont de
 * Dinechin "Parametric Computation of Margins and of Minimum Cumulative Register
 * Lifetime Dates", LCPC'96 --	9th International Workshop on Languages and
 * Compilers for Parallel Computing, LNCS #1239.
 */
void
DDGraph_addRegisterRAW(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                       int theta, int omega, Temporary temporary, bool lifeOK, int relaxable);

/*
 * DDGraph_addRegisterWAR --	Add a WAR register dependence arc between two DDGraphNodes.
 *
 * The arc is omitted in case it is a self-arc that is guaranteed non-critical.
 */
void
DDGraph_addRegisterWAR(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                       int theta, int omega, Temporary temporary);

/*
 * DDGraph_addRegisterWAW --	Add a WAW register dependence arc between two DDGraphNodes.
 *
 * The arc is omitted in case it is a self-arc that is guaranteed non-critical.
 * If =lifeOK=, a =life_node= is created and a Life arc is added between
 * =head_node= and =life_node=. This has the effect of pulling =tail_node= and
 * =head_node= together. Even though both nodes write into the same register (WAW),
 * if they belong to the same Lifetime then =tail_node= does not kill the
 * register. In such case it makes sense to pull the nodes together.
 */
void
DDGraph_addRegisterWAW(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                       int theta, int omega, Temporary temporary, bool lifeOK, bool isKill);

/*
 * DDGraph_addRegionFlow --	Add a Flow control dependence between two DDGraphNodes.
 *
 * The arc is omitted in case it is a self-arc that is guaranteed non-critical.
 */
void
DDGraph_addRegionFlow(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                      int theta, int omega, Temporary temporary);

/*
 * DDGraph_addControlOutput --	Add a Output control dependence between two DDGraphNodes.
 *
 * The arc is omitted in case it is a self-arc that is guaranteed non-critical.
 */
void
DDGraph_addControlOutput(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                         int theta, int omega, Temporary temporary);

/*
 * DDGraph_addControlAnti --	Add a Anti control dependence between two DDGraphNodes.
 *
 * The arc is omitted in case it is a self-arc that is guaranteed non-critical.
 */
void
DDGraph_addControlAnti(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                       int theta, int omega, Temporary temporary);

/*
 * DDGraph_addMemoryArc --	Add a memory dependence arc between two DDGraphNodes.
 */
void
DDGraph_addMemoryArc(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                     int theta, int omega, Temporary temporary, DependenceKind kind);

/*
 * DDGraph_addSequenceArc --	Add a sequence arc.
 */
void
DDGraph_addSequenceArc(DDGraph this, DDGraphNode tail_node, DDGraphNode head_node,
                       int theta, int omega, Temporary temporary);

/*
 * DDGraph_addRegisterArcs --	Dependency computation for register resources.
 *
 * The outer loop iterates over the Write effects (effect1), and the two inner
 * loops that over the use effects (effect2 and effect4).
 */
void
DDGraph_addRegisterArcs(DDGraph this, Lifetime lifetime,
                        bool isAcyclic, bool isPrePass, int l1MissExtra);

/*
 * DDGraph_addControlArcs --	Dependency computation for control resources.
 *
 * The main differences with the Register resources are: control speculation is
 * enabled by removing some control dependences, depending on the value of
 * speculation; and anti dependences are not enforced if compensation is allowed
 * because an instruction can be pushed below an exit branch.
 */
void
DDGraph_addControlArcs(DDGraph this, Lifetime lifetime, bool isAcyclic);

/*
 * DDGraph_addMemoryArcs --	Dependency computation for the memory resources.
 *
 * The computation is similar to DDGraph_addRegisterArcs.
 */
void
DDGraph_addMemoryArcs(DDGraph this, Lifetime lifetime, bool isAcyclic, const_Configure configure);

/*
 * DDGraph_addDependences --	Add the LoopScope Dependences to this DDGraph.
 */
void
DDGraph_addDependences(DDGraph this,
                       PtrSet ddgNodes, const_DependenceTable dependenceTable,
                       Temporary temporary, bool isAcyclic, const_Configure configure);

/*
 * DDGraph_addSPUpdateArcs --	Add dependences between SP update and memory operations.
 */
void
DDGraph_addSPUpdateArcs(DDGraph this, BlockTrace blockTrace,
                        bool isAcyclic, Register stackRegister);

/*
 * DDGraph_addVolatileArcs --	Add the volatile arcs.
 */
void
DDGraph_addVolatileArcs(DDGraph this, Lifetime lifetime, bool isAcyclic);

/*
 * DDGraph_INIT --	Initialize the DDGraph module.
 */
void
DDGraph_INIT(void);

/*
 * DDGraph_FINI --	Finalize the DDGraph module.
 */
#define DDGraph_FINI()


#endif//DDGraph__h_INCLUDED
