#ifndef String_h_INCLUDED
#define String_h_INCLUDED

/*
 * !!!!	String.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 *
 * An String is a single-copy, non-mutable object, that can contain arbitrary data.
 * String(s) pointers are equal only when the underlying data is the same. The
 * single copy property of String(s) is only relative to a particular StringTable (String
 * Table). The StringTable itself organizes its String(s) into a balanced binary tree
 * search structure.
 */

/*
 * String --	A single-copy, non-mutable object, maintained in a StringTable.
 *
 * An String is aligned to =intptr_t= alignment constraints, and its storage is
 * rounded up to =sizeof(intptr_t)=. The leftover storage bytes are cleared.
 */
typedef const char *String;

/*
 * String_ENTRY --	Access the underlying StringTableEntry of this String.
 */
#define String_ENTRY(this) (((StringTableEntry)((void*)(this))) - 1)

/*
 * String_Size --	The memory size needed by this String
 */
#define String_Size(data, length, leaf) ( \
  StringTableEntry_DATA(leaf) != NULL? \
      ((length) + sizeof(intptr_t) - 1) & -sizeof(intptr_t): 0 \
)

/*
 * StringTableEntry --	Information to manage a String into a StringTable.
 */
struct StringTableEntry_ {
  //@args	const void *data, int32_t length, StringTableEntry leaf
  struct StringTableEntry_ *LEFT;	// The left kid in the binary search tree.
  struct StringTableEntry_ *RIGHT;	// The right kid in the binary search tree.
  uint32_t HASHED;	// The hashed data.
  //@access PRIORITY	(StringTableEntry_HASHED(this))
  int32_t LENGTH;	// The data length.
  const void *DATA;	// Where the data is located.
  //@access STRING	((String)((StringTableEntry)(this) + 1))
};
typedef struct StringTableEntry_ StringTableEntry_, *StringTableEntry;
typedef const struct StringTableEntry_ *const_StringTableEntry;
typedef struct StringTableEntry_ * restrict_StringTableEntry;

extern StringTableEntry
StringTableEntry_Ctor(StringTableEntry this, const void *data, int32_t length, StringTableEntry leaf);

extern StringTableEntry
StringTableEntry_Copy(StringTableEntry this, const_StringTableEntry that);

extern void
StringTableEntry_Dtor(StringTableEntry this);

extern size_t
StringTableEntry_Size(const void *data, int32_t length, StringTableEntry leaf);

#define StringTableEntry_LEFT(this) ((this)->LEFT)
#define StringTableEntry__LEFT(this) (&(this)->LEFT)
#define StringTableEntry_RIGHT(this) ((this)->RIGHT)
#define StringTableEntry__RIGHT(this) (&(this)->RIGHT)
#define StringTableEntry_HASHED(this) ((this)->HASHED)
#define StringTableEntry__HASHED(this) (&(this)->HASHED)
#define StringTableEntry_LENGTH(this) ((this)->LENGTH)
#define StringTableEntry__LENGTH(this) (&(this)->LENGTH)
#define StringTableEntry_DATA(this) ((this)->DATA)
#define StringTableEntry__DATA(this) (&(this)->DATA)
#define StringTableEntry_PRIORITY(this) ((StringTableEntry_HASHED(this)))
#define StringTableEntry_STRING(this) (((String)((StringTableEntry)(this) + 1)))


//
/*
 * String_hashed --	return the hashed value of =this= String.
 */
static inline uint32_t
String_hashed(String this)
{
  StringTableEntry entry = String_ENTRY(this);
  return StringTableEntry_HASHED(entry);
}

/*
 * String_length --	return the length of =this= String.
 */
static inline int32_t
String_length(String this)
{
  StringTableEntry entry = String_ENTRY(this);
  return StringTableEntry_LENGTH(entry);
}

/*
 * String_data --	return the data of =this= String.
 */
static inline const void *
String_data(String this)
{
  StringTableEntry entry = String_ENTRY(this);
  return StringTableEntry_DATA(entry);
}

/*
 * StringTable --	Implements a table of String(s) without duplicates.
 *
 * =copies=:	If set, the string will be copied when entered in the StringTable.
 */
struct StringTable_ {
  //@args	Memory memory, bool copies
  Memory MEMORY;	// Where the String(s) are allocated.
  struct StringTableEntry_ *ROOT;
  StringTableEntry_ LEAF_;
  //@access LEAF	StringTable__LEAF_(this)
  //@access COPIES	(StringTableEntryDATA(StringTable_LEAF(this)) == (const void *)1)
  uint32_t COUNT;
};
typedef struct StringTable_ StringTable_, *StringTable;
typedef const struct StringTable_ *const_StringTable;
typedef struct StringTable_ * restrict_StringTable;

extern StringTable
StringTable_Ctor(StringTable this, Memory memory, bool copies);

extern StringTable
StringTable_Copy(StringTable this, const_StringTable that);

extern void
StringTable_Dtor(StringTable this);

extern size_t
StringTable_Size(Memory memory, bool copies);

#define StringTable_MEMORY(this) ((this)->MEMORY)
#define StringTable__MEMORY(this) (&(this)->MEMORY)
#define StringTable_ROOT(this) ((this)->ROOT)
#define StringTable__ROOT(this) (&(this)->ROOT)
#define StringTable_LEAF_(this) ((this)->LEAF_)
#define StringTable__LEAF_(this) (&(this)->LEAF_)
#define StringTable_COUNT(this) ((this)->COUNT)
#define StringTable__COUNT(this) (&(this)->COUNT)
#define StringTable_COPIES(this) ((StringTableEntryDATA(StringTable_LEAF(this)) == (const void *)1))
#define StringTable_LEAF(this) (StringTable__LEAF_(this))


/*
  StringTable_memory -- This StringTable Memory.
 */
static inline Memory
StringTable_memory(const_StringTable this)
{
  return StringTable_MEMORY(this);
}

/*
 */
static inline bool
StringTable_copies(const_StringTable this)
{
  const_StringTableEntry leaf = StringTable_LEAF(this);
  return StringTableEntry_DATA(leaf) != NULL;
}

/*
 * StringTable_search --	Search the String corresponding to data.
 *
 * Return:	The String corresponding to data, or NULL if not found.
 */
String
StringTable_search(const_StringTable this, const void *data, int32_t length);

/*
 * StringTable_insert --	Insert data into this table, unless it's already there.
 */
String
StringTable_insert(StringTable this, const void *data, int32_t length);

/*
 * StringTable_remove --	Remove data from the tree if it's there.
 */
bool
StringTable_remove(StringTable this, const void *data, int32_t length);

/*
 * StringTable_map --	Map the StringTable String(s) using the supplied =map= function.
 * map:	The function that will be called for each String in the StringTable.
 *               First argument of =(*map)= is the current String. Second argument
 *               of =(*map)= is the =va_list= corresponding to the =...=
 *               arguments of StringTable_map.
 *  =...=:	The other arguments if any to be passed to =(*map)= as =va_list=.
 */
void
StringTable_map(StringTable this, void (*map)(String, va_list), ...);

/*
 * String_S --	Make a string from %s.
 */
String
StringTable_S(StringTable this, const char *name);

/*
 * String_D --	Make a string from %lld.
 */
String
StringTable_D(StringTable this, int64_t lld);

/*
 * StringStoreItem --	Information attached to a String in a StringStore.
 */
struct StringStoreItem_ {
  //@args	int32_t length, struct StringStoreItem_ *next
  int32_t LENGTH;
  struct StringStoreItem_ *NEXT;
};
typedef struct StringStoreItem_ StringStoreItem_, *StringStoreItem;
typedef const struct StringStoreItem_ *const_StringStoreItem;
typedef struct StringStoreItem_ * restrict_StringStoreItem;

extern StringStoreItem
StringStoreItem_Ctor(StringStoreItem this, int32_t length, struct StringStoreItem_ *next);

extern StringStoreItem
StringStoreItem_Copy(StringStoreItem this, const_StringStoreItem that);

extern void
StringStoreItem_Dtor(StringStoreItem this);

extern size_t
StringStoreItem_Size(int32_t length, struct StringStoreItem_ *next);

#define StringStoreItem_LENGTH(this) ((this)->LENGTH)
#define StringStoreItem__LENGTH(this) (&(this)->LENGTH)
#define StringStoreItem_NEXT(this) ((this)->NEXT)
#define StringStoreItem__NEXT(this) (&(this)->NEXT)


/*
 * StringStore --	Simple store to save strings.
 */
struct StringStore_ {
  //@args	Memory memory
  Memory MEMORY;
  StringStoreItem FIRST;
};
typedef struct StringStore_ StringStore_, *StringStore;
typedef const struct StringStore_ *const_StringStore;
typedef struct StringStore_ * restrict_StringStore;

extern StringStore
StringStore_Ctor(StringStore this, Memory memory);

extern StringStore
StringStore_Copy(StringStore this, const_StringStore that);

extern void
StringStore_Dtor(StringStore this);

extern size_t
StringStore_Size(Memory memory);

#define StringStore_MEMORY(this) ((this)->MEMORY)
#define StringStore__MEMORY(this) (&(this)->MEMORY)
#define StringStore_FIRST(this) ((this)->FIRST)
#define StringStore__FIRST(this) (&(this)->FIRST)


/*
 * StringStore_memory --	This StringStore Memory.
 */
static inline Memory
StringStore_memory(const_StringStore this)
{
  return StringStore_MEMORY(this);
}

/*
 * StringStore_first --	For use in StringStore_FOREACH_String.
 */
static inline StringStoreItem
StringStore_first(const_StringStore this)
{
  return StringStore_FIRST(this);
}

/*
 * StringStore_FOREACH_String --	Iterate this StringStore String(s).
 */
#define StringStore_FOREACH_String(this, string) { \
  StringStoreItem StringStore_ITEM = StringStore_first(this); \
  while (StringStore_ITEM != NULL) { \
    StringStoreItem StringStore_NEXT = StringStoreItem_NEXT(StringStore_ITEM); \
    String string = (String)(StringStore_ITEM + 1);
#define StringStore_ENDEACH_String \
    StringStore_ITEM = StringStore_NEXT; \
  } \
}

/*
 * StringStore_push --	Push a String onto this StringStore.
 */
String
StringStore_push(StringStore this, const void *data, int32_t length);

/*
 * StringStore_S --	Push a null-teminated C string onto this StringStore.
 */
static inline String
StringStore_S(StringStore this, const char *name)
{
  int32_t length = strlen(name) + 1;
  return StringStore_push(this, name, length);
}


#endif//String_h_INCLUDED
