#ifndef Hierarchy_h_INCLUDED
#define Hierarchy_h_INCLUDED

/*
 * !!!!	Hierarchy.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * HierarchyComponentType --	Enumerate the HierarchyComponent types.
 */
typedef enum {
  HierarchyComponentType_Node,
  HierarchyComponentType_Root,
  HierarchyComponentType_SCC,
  HierarchyComponentType_Loop,
  HierarchyComponentType__
} HierarchyComponentType;
typedef uint8_t short_HierarchyComponentType;

/*
 * HierarchyComponent --	Region in a Hierarchy of a DGraph.
 */
struct HierarchyComponent_ {
  //@args	HierarchyComponentType type, DGraphNode node
  short_HierarchyComponentType TYPE;	// This HierarchyComponent type.
  int16_t DEPTH;			// Nesting level of this HierarchyComponent.
  DGraphNode NODE;			// DGraphNode.
  struct HierarchyComponent_ *CHILD;	// First child of this HierarchyComponent.
  struct HierarchyComponent_ *NEXT;	// Next child in parent HierarchyComponent.
};
typedef struct HierarchyComponent_ HierarchyComponent_, *HierarchyComponent;
typedef const struct HierarchyComponent_ *const_HierarchyComponent;
typedef struct HierarchyComponent_ * restrict_HierarchyComponent;

extern HierarchyComponent
HierarchyComponent_Ctor(HierarchyComponent this, HierarchyComponentType type, DGraphNode node);

extern HierarchyComponent
HierarchyComponent_Copy(HierarchyComponent this, const_HierarchyComponent that);

extern void
HierarchyComponent_Dtor(HierarchyComponent this);

extern size_t
HierarchyComponent_Size(HierarchyComponentType type, DGraphNode node);

#define HierarchyComponent_TYPE(this) ((this)->TYPE)
#define HierarchyComponent__TYPE(this) (&(this)->TYPE)
#define HierarchyComponent_DEPTH(this) ((this)->DEPTH)
#define HierarchyComponent__DEPTH(this) (&(this)->DEPTH)
#define HierarchyComponent_NODE(this) ((this)->NODE)
#define HierarchyComponent__NODE(this) (&(this)->NODE)
#define HierarchyComponent_CHILD(this) ((this)->CHILD)
#define HierarchyComponent__CHILD(this) (&(this)->CHILD)
#define HierarchyComponent_NEXT(this) ((this)->NEXT)
#define HierarchyComponent__NEXT(this) (&(this)->NEXT)


//
static inline HierarchyComponentType
HierarchyComponent_type(const_HierarchyComponent this)
{
  return HierarchyComponent_TYPE(this);
}

//
static inline int
HierarchyComponent_depth(const_HierarchyComponent this)
{
  return HierarchyComponent_DEPTH(this);
}

//
static inline DGraphNode
HierarchyComponent_node(const_HierarchyComponent this)
{
  return HierarchyComponent_NODE(this);
}

//
static inline HierarchyComponent
HierarchyComponent_child(const_HierarchyComponent this)
{
  return HierarchyComponent_CHILD(this);
}

//
static inline HierarchyComponent
HierarchyComponent_next(const_HierarchyComponent this)
{
  return HierarchyComponent_NEXT(this);
}

//
#define HierarchyComponent_CHILDREN_FOREACH_HierarchyComponent(this, component) { \
  HierarchyComponent component = HierarchyComponent_child(this); \
  for (; component != NULL; component = HierarchyComponent_next(component)) {
#define HierarchyComponent_CHILDREN_ENDEACH_HierarchyComponent \
  } \
}

//
bool
HierarchyComponent_pretty(const_HierarchyComponent this, FILE *file,
                          bool (*node_pretty)(DGraphNode, FILE*));

//
void
HierarchyComponent_free(HierarchyComponent this, Memory memory);

/*
 * Hierarchy --	Hierarchy of Strongly Connected Regions.
 */
struct Hierarchy_ {
  //@args	Memory memory, DGraph dgraph, DGraphNode entry
  Memory MEMORY;
  int32_t COUNT;
  DGraphNode *NODEOF;		// Array of DGraphNode(s) indexed by DGraphNode_NUMBER.
  DGraphNode ENTRY;
  int32_t *DFN;			// Array of DFN numbers indexed by DGraphNode_NUMBER.
  int32_t *PAST;		// Past the stack pointer.
  HierarchyComponent_ ROOT_;
  //@access ROOTNODE	Hierarchy__ROOT_(this)
};
typedef struct Hierarchy_ Hierarchy_, *Hierarchy;
typedef const struct Hierarchy_ *const_Hierarchy;
typedef struct Hierarchy_ * restrict_Hierarchy;

extern Hierarchy
Hierarchy_Ctor(Hierarchy this, Memory memory, DGraph dgraph, DGraphNode entry);

extern Hierarchy
Hierarchy_Copy(Hierarchy this, const_Hierarchy that);

extern void
Hierarchy_Dtor(Hierarchy this);

extern size_t
Hierarchy_Size(Memory memory, DGraph dgraph, DGraphNode entry);

#define Hierarchy_MEMORY(this) ((this)->MEMORY)
#define Hierarchy__MEMORY(this) (&(this)->MEMORY)
#define Hierarchy_COUNT(this) ((this)->COUNT)
#define Hierarchy__COUNT(this) (&(this)->COUNT)
#define Hierarchy_NODEOF(this) ((this)->NODEOF)
#define Hierarchy__NODEOF(this) (&(this)->NODEOF)
#define Hierarchy_ENTRY(this) ((this)->ENTRY)
#define Hierarchy__ENTRY(this) (&(this)->ENTRY)
#define Hierarchy_DFN(this) ((this)->DFN)
#define Hierarchy__DFN(this) (&(this)->DFN)
#define Hierarchy_PAST(this) ((this)->PAST)
#define Hierarchy__PAST(this) (&(this)->PAST)
#define Hierarchy_ROOT_(this) ((this)->ROOT_)
#define Hierarchy__ROOT_(this) (&(this)->ROOT_)
#define Hierarchy_ROOTNODE(this) (Hierarchy__ROOT_(this))


//
Hierarchy
Hierarchy_new(Memory parent, DGraph dgraph, DGraphNode entry);

//
Hierarchy
Hierarchy_delete(Hierarchy this);

//
static inline HierarchyComponent
Hierarchy_root(Hierarchy this)
{
  return Hierarchy_ROOTNODE(this);
}

//
bool
Hierarchy_pretty(const_Hierarchy this, FILE *file,
                 bool (*node_pretty)(DGraphNode, FILE*));

/*
 * Hierarchy_makeSCC --	Make a Hierarchy of Strongly Connected Components.
 */
void
Hierarchy_makeSCC(Hierarchy this);

/*
 * Hierarchy_makeNSCC --	Make a Hierarchy of Nested Strongly Connected Components.
 *
 * The Nested Strongly Connected Components is a Loop Forest as defined by
 * G. Ramalingam, ``Identifying loops in almost linear time'', Transactions on
 * Programming Languages and Systems, Volume 21, Number 2, March 1999. F.
 * Bourdoncle introduced this decomposition in his PhD thesis in order to
 * minimize the number of widening points in a control-flow graph.
 */
void
Hierarchy_makeNSCC(Hierarchy this);


#endif//Hierarchy_h_INCLUDED
