#ifndef Temporary_h_INCLUDED
#define Temporary_h_INCLUDED

/*
 * !!!!	Temporary.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 * Christophe Guillon (Christophe.Guillon@st.com)
 * Nikola Puzovic (Nikola.Puzovic@st.com)
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 *  !!!	Temporary Types
 *
 *   !!	Dedicated Temporary(ies)
 *
 * Dedicated Temporary(ies) are immutable and used to represent Operation arguments
 * or results that must be in specific machine registers. This applies to procedure
 * register arguments and results an also to implicit or pinned register operands.
 * In effect, using a Dedicated Temporary is an override for the register allocator
 * on what Register to assign to a particular Temporary.
 *
 * Temporary_hasRegFile is true for Dedicated Temporary(ies)
 * Temporary_hasRegister is true for Dedicated Temporary(ies)
 *
 * See the note below for more details on dedicated temporary usage.
 *
 *   !!	Virtual Temporary(ies)
 *
 * Virtual Temporary(ies) that are also known as pseudo-registers. They are used to
 * carry values between Operation(s).
 *
 * Temporary_hasRegFile is true for Virtual Temporary(ies)
 * Temporary_hasRegister is valid for Virtual Temporary(ies)
 *
 *   !!	Variable Temporary(ies)
 *
 * Variable Temporary(ies) are base objects of Variable(s). They can be assigned a Register.
 *
 * Temporary_hasRegFile is true for Variable Temporary(ies)
 * Temporary_hasRegister is valid for Variable Temporary(ies)
 *
 *   !!	Absolute Temporary(ies)
 *
 * Absolute Temporary(ies) are used to represent resolved constants.
 *
 * Temporary_isInvariant is true for Absolute Temporary(ies)
 *
 *   !!	Label Temporary(ies)
 *
 * Label Temporary(ies) are used for the arguments for intra-procedural branches. For
 * inter-procedural branches, such as those that result from optimizing tail calls,
 * including recursive tail call, Symbol Temporary(ies) must be used instead.
 *
 * Temporary_isInvariant is true for Label Temporary(ies)
 *
 *   !!	Symbol Temporary(ies)
 *
 * Symbol Temporary(ies) are used to represent symbolic values made of a base Symbol
 * plus an offset.
 *
 * Temporary_isInvariant is true for Symbol Temporary(ies)
 *
 *   !!	Modifier Temporary(ies)
 *
 * Modifier Temporary(ies) are used to represent instruction modifiers.
 *
 * Temporary_isInvariant is true for Modifier Temporary(ies)
 *
 *   !!	RegMask Temporary(ies)
 *
 * RegMask Temporary(ies) are placeholders for the immdiate values interpreted as
 * register masks.  During encoding, the value of the Instruction's register mask
 * is computed from the Operation's additional arguments or results that are not
 * Parameters.
 *
 * Temporary_isInvariant is true for RegMask Temporary(ies)
 *
 *   !!	Note on dedicated Temporary(ies)
 *
 * It is a requirement for correctness to use dedicated temporaries in the following cases:
 * *	use of procedure incoming parameters in and only in the prologue basic block
 * (the procedure entry),
 * *	definition of procedure outgoing results in and only in any epilogue basic
 * block containing a exit from the procedure,
 * *	definition of function call parameter in and only in the block where the call
 * site appears,
 * *	use of function result in and only in the block immediatly following the block
 * where the call site appears, this block must exist and have a single predecessor
 * as soon as some return value of the call site is used,
 * *	any use or definition of defined registers such as the stack pointer, the
 * frame pointer, the global data pointer and the procedure link register must use
 * the corresponding dedicated temporary.
 *
 * These requirements must be enforced before register allocation. After register
 * allocation, the same requirements hold but the scope of the dedicated temporaries
 * used for parameter passing can be extended out of the homing basic block.
 *
 *   !!	Note on sharing Temporary(ies)
 *
 * The Register Temporary(ies) should be shared between Operation(s).
 * Other Temporary(ies) should be used by only one Operation. This can be relaxed
 * for Temporary(ies) that are immutable like Absolute Temporary(ies), and also in
 * case it is guaranteed all Operation(s) that share a non-Register Temporary do
 * not update it. Updates are for instance changing the Relocation of the offset.
 *
 *   !!	Note on Temporary Relocation(s)
 *
 * Invariant Temporary(ies) have an associated Relocation. This Relocation is an
 * enumeration that specifies additional processing or checking to be done on a
 * resolved ImmediateValue before being passed to the Immediate encoding. Thus,
 * processing done by a linker relocation combines the processing of Relocation
 * with the Immediate encoding process.
 */

/*
 * Temporary_Factory --	IndexedFactory for Temporary(ies).
 */
extern IndexedFactory_ Temporary_Factory[1];

/*
 * TemporaryType --	Enumerates the Temporary types.
 */
typedef enum {
  TemporaryType_Dedicated,		// Dedicated register.
                                        //	REGISTER is the Register.
                                        //	WIDTH is the bit-width if known.
  TemporaryType_Virtual,		// Virtual register.
                                        //	REGISTER is the first of RegFile.
                                        //	WIDTH is the bit-width if known.
  TemporaryType_Variable,		// Variable base.
                                        //	REGISTER is the first of RegFile.
                                        //	WIDTH is the bit-width if known.
  TemporaryType_Absolute,		// Absolute invariant.
                                        //	INVARIANT is the TemporaryInvariant.
  TemporaryType_Symbol,			// Symbol invariant.
                                        //	INVARIANT is the TemporaryInvariant.
  TemporaryType_Label,			// Label invariant.
                                        //	INVARIANT is the TemporaryInvariant.
  TemporaryType_Modifier,               // Modifier invariant.
                                        //	INVARIANT is the TemporaryInvariant.
  TemporaryType_RegMask,		// RegMask invariant.
  TemporaryType_Volatile,		// Volatile resource.
  TemporaryType_Control,		// Control resource.
  TemporaryType_Memory,			// Memory resource.
  TemporaryType__
} enum_TemporaryType;
typedef uint8_t TemporaryType;
#define TemporaryType_isDedicated(type)	((type) == TemporaryType_Dedicated)
#define TemporaryType_isVirtual(type)	((type) == TemporaryType_Virtual)
#define TemporaryType_isVariable(type)	((type) == TemporaryType_Variable)
#define TemporaryType_isAbsolute(type)	((type) == TemporaryType_Absolute)
#define TemporaryType_isSymbol(type)	((type) == TemporaryType_Symbol)
#define TemporaryType_isLabel(type)	((type) == TemporaryType_Label)
#define TemporaryType_isModifier(type)	((type) == TemporaryType_Modifier)
#define TemporaryType_isRegMask(type)	((type) == TemporaryType_RegMask)
#define TemporaryType_isVolatile(type)	((type) == TemporaryType_Volatile)
#define TemporaryType_isControl(type)	((type) == TemporaryType_Control)
#define TemporaryType_isMemory(type)	((type) == TemporaryType_Memory)
#define TemporaryType_hasRegFile(type)	\
  ((unsigned)(type) <= TemporaryType_Variable)
#define TemporaryType_isInvariant(type)	\
  (   (unsigned)((type) - TemporaryType_Absolute) \
   <= (unsigned)(TemporaryType_RegMask - TemporaryType_Absolute))
#define TemporaryType_isRelaxable(type)	\
  (   (unsigned)((type) - TemporaryType_Absolute) \
   <= (unsigned)(TemporaryType_Symbol - TemporaryType_Absolute))
extern const char *
TemporaryType_name_(TemporaryType this);

/*
 * TemporaryFlag --	Enumerates the Temporary flags.
 */
typedef enum {
  TemporaryFlag_Constant = 0x1,		// Temporary is constant.
  TemporaryFlag_Register = 0x2,		// Temporary has a Register.
  TemporaryFlag_Global = 0x4,		// Temporary is global in Liveness.
  TemporaryFlag_Killed = 0x8,		// Liveness flag, must be cleared after use.
  TemporaryFlag_Remater = 0x10,		// Register Temporary is rematerializable.
                                        //	Get Invariant Temporary in REMATERMAP.
  TemporaryFlag_Homeable = 0x20,	// Register Temporary is homeable.
                                        //	Get home Symbol Temporary in HOMEABLEMAP.
  TemporaryFlag_RenameOK = 0x40,	// Virtual Temporary can be SSA renamed.
  TemporaryFlag_Recycled = 0x80,	// Virtual Temporary is recycled by SSADestruct.
} enum_TemporaryFlag;
typedef uint8_t TemporaryFlags;

//
extern struct ETable_ TemporaryFlag_Table[1];

/*
 * TemporaryUnion_ --	Union for the variant part of Temporary.
 */
union TemporaryUnion_ {
  struct Variable_ *WEBROOT;		// TemporaryType_isVirtual.
  struct Variable_ *SPILLED;		// TemporaryType_isVariable.
};
typedef union TemporaryUnion_ TemporaryUnion_, *TemporaryUnion;

/*
 * Temporary --	Argument and result type for Operations.
 */
struct Temporary_ {
  //@args	TemporaryType type, unsigned flags
  Indexed_ __;				// A Temporary_ ISA Indexed_.
  //@access IDENTITY	Indexed_IDENTITY(Temporary____(this))
  //@access INDEX	Indexed_INDEX(Temporary____(this))
  //@mutate INDEX	Indexed__INDEX(Temporary____(this))
  TemporaryUnion_ UNION_;		// See TemporaryUnion.
  //@access WEBROOT	(Temporary_UNION_(this).WEBROOT)
  //@mutate WEBROOT	(&Temporary_UNION_(this).WEBROOT)
  //@access SPILLED	(Temporary_UNION_(this).SPILLED)
  //@mutate SPILLED	(&Temporary_UNION_(this).SPILLED)
  TemporaryType TYPE;			// This Temporary type.
  //@access isDedicated	TemporaryType_isDedicated(Temporary_type(this))
  //@access isVirtual	TemporaryType_isVirtual(Temporary_type(this))
  //@access isVariable	TemporaryType_isVariable(Temporary_type(this))
  //@access isAbsolute	TemporaryType_isAbsolute(Temporary_type(this))
  //@access isSymbol	TemporaryType_isSymbol(Temporary_type(this))
  //@access isLabel	TemporaryType_isLabel(Temporary_type(this))
  //@access isModifier	TemporaryType_isModifier(Temporary_type(this))
  //@access isRegMask	TemporaryType_isRegMask(Temporary_type(this))
  //@access isVolatile	TemporaryType_isVolatile(Temporary_type(this))
  //@access isControl	TemporaryType_isControl(Temporary_type(this))
  //@access isMemory	TemporaryType_isMemory(Temporary_type(this))
  //@access hasRegFile	TemporaryType_hasRegFile(Temporary_type(this))
  //@access isInvariant	TemporaryType_isInvariant(Temporary_type(this))
  //@access isRelaxable	TemporaryType_isRelaxable(Temporary_type(this))
  TemporaryFlags FLAGS;		// This Temporary flags.
  //@access isConstant	((Temporary_flags(this) & TemporaryFlag_Constant) != 0)
  //@access hasRegister	((Temporary_flags(this) & TemporaryFlag_Register) != 0)
  //@access isGlobal	((Temporary_flags(this) & TemporaryFlag_Global) != 0)
  //@access isKilled	((Temporary_flags(this) & TemporaryFlag_Killed) != 0)
  //@access isRemater	((Temporary_flags(this) & TemporaryFlag_Remater) != 0)
  //@access isHomeable	((Temporary_flags(this) & TemporaryFlag_Homeable) != 0)
  //@access isRenameOK	((Temporary_flags(this) & TemporaryFlag_RenameOK) != 0)
  //@access isRecycled	((Temporary_flags(this) & TemporaryFlag_Recycled) != 0)
  //@access isAssigned	(Temporary_isVirtual(this) && Temporary_hasRegister(this))
  Register REGISTER;			// The Register of this Temporary.
  //@access REGFILE	(Register_regFile(Temporary_REGISTER(this)))
  uint8_t WIDTH;			// Width in bits of this Temporary.
  //@access isRenameReg	(Temporary_hasRegFile(this) && !Temporary_isDedicated(this))
  //@access INVARIANT	(TemporaryInvariant)(this)
  //@access VARIABLE	(Variable)(this)
};
typedef struct Temporary_ Temporary_, *Temporary;
typedef const struct Temporary_ *const_Temporary;
typedef struct Temporary_ * restrict_Temporary;

extern Temporary
Temporary_Ctor(Temporary this, TemporaryType type, unsigned flags);

extern Temporary
Temporary_Copy(Temporary this, const_Temporary that);

extern void
Temporary_Dtor(Temporary this);

extern size_t
Temporary_Size(TemporaryType type, unsigned flags);

#define Temporary___(this) ((this)->__)
#define Temporary____(this) (&(this)->__)
#define Temporary_UNION_(this) ((this)->UNION_)
#define Temporary__UNION_(this) (&(this)->UNION_)
#define Temporary_TYPE(this) ((this)->TYPE)
#define Temporary__TYPE(this) (&(this)->TYPE)
#define Temporary_FLAGS(this) ((this)->FLAGS)
#define Temporary__FLAGS(this) (&(this)->FLAGS)
#define Temporary_REGISTER(this) ((this)->REGISTER)
#define Temporary__REGISTER(this) (&(this)->REGISTER)
#define Temporary_WIDTH(this) ((this)->WIDTH)
#define Temporary__WIDTH(this) (&(this)->WIDTH)
#define Temporary_isRecycled(this) (((Temporary_flags(this) & TemporaryFlag_Recycled) != 0))
#define Temporary_isConstant(this) (((Temporary_flags(this) & TemporaryFlag_Constant) != 0))
#define Temporary_hasRegFile(this) (TemporaryType_hasRegFile(Temporary_type(this)))
#define Temporary_isControl(this) (TemporaryType_isControl(Temporary_type(this)))
#define Temporary_isVariable(this) (TemporaryType_isVariable(Temporary_type(this)))
#define Temporary_isRegMask(this) (TemporaryType_isRegMask(Temporary_type(this)))
#define Temporary_SPILLED(this) ((Temporary_UNION_(this).SPILLED))
#define Temporary_VARIABLE(this) ((Variable)(this))
#define Temporary_isVolatile(this) (TemporaryType_isVolatile(Temporary_type(this)))
#define Temporary_isKilled(this) (((Temporary_flags(this) & TemporaryFlag_Killed) != 0))
#define Temporary_hasRegister(this) (((Temporary_flags(this) & TemporaryFlag_Register) != 0))
#define Temporary_isAssigned(this) ((Temporary_isVirtual(this) && Temporary_hasRegister(this)))
#define Temporary_isHomeable(this) (((Temporary_flags(this) & TemporaryFlag_Homeable) != 0))
#define Temporary_isModifier(this) (TemporaryType_isModifier(Temporary_type(this)))
#define Temporary_isRelaxable(this) (TemporaryType_isRelaxable(Temporary_type(this)))
#define Temporary_INDEX(this) (Indexed_INDEX(Temporary____(this)))
#define Temporary_isRenameOK(this) (((Temporary_flags(this) & TemporaryFlag_RenameOK) != 0))
#define Temporary_WEBROOT(this) ((Temporary_UNION_(this).WEBROOT))
#define Temporary_isDedicated(this) (TemporaryType_isDedicated(Temporary_type(this)))
#define Temporary_isAbsolute(this) (TemporaryType_isAbsolute(Temporary_type(this)))
#define Temporary_isSymbol(this) (TemporaryType_isSymbol(Temporary_type(this)))
#define Temporary_isRenameReg(this) ((Temporary_hasRegFile(this) && !Temporary_isDedicated(this)))
#define Temporary_isGlobal(this) (((Temporary_flags(this) & TemporaryFlag_Global) != 0))
#define Temporary_isMemory(this) (TemporaryType_isMemory(Temporary_type(this)))
#define Temporary_INVARIANT(this) ((TemporaryInvariant)(this))
#define Temporary_isLabel(this) (TemporaryType_isLabel(Temporary_type(this)))
#define Temporary_REGFILE(this) ((Register_regFile(Temporary_REGISTER(this))))
#define Temporary_IDENTITY(this) (Indexed_IDENTITY(Temporary____(this)))
#define Temporary_isVirtual(this) (TemporaryType_isVirtual(Temporary_type(this)))
#define Temporary_isRemater(this) (((Temporary_flags(this) & TemporaryFlag_Remater) != 0))
#define Temporary_isInvariant(this) (TemporaryType_isInvariant(Temporary_type(this)))
#define Temporary__INDEX(this) (Indexed__INDEX(Temporary____(this)))
#define Temporary__SPILLED(this) ((&Temporary_UNION_(this).SPILLED))
#define Temporary__WEBROOT(this) ((&Temporary_UNION_(this).WEBROOT))


//
static inline Indexed
Temporary__(Temporary this)
{
  return Temporary____(this);
}
static inline const_Indexed
Temporary__const(const_Temporary this)
{
  return Temporary____(this);
}

/*
 * Temporary_identity --	This Temporary identity.
 */
static inline uint32_t
Temporary_identity(const_Temporary this)
{
  return Temporary_IDENTITY(this);
}

/*
 * Temporary_index --	This Temporary index.
 */
static inline uint32_t
Temporary_index(const_Temporary this)
{
  return Temporary_INDEX(this);
}

/*
 * Temporary_type --	This Temporary TemporaryType.
 */
static inline TemporaryType
Temporary_type(const_Temporary this)
{
  return Temporary_TYPE(this);
}

/*
 * Temporary_flags --	This Temporary TemporaryFlags.
 */
static inline TemporaryFlags
Temporary_flags(const_Temporary this)
{
  return Temporary_FLAGS(this);
}

/*
 * Temporary_setFlags --	Set this Temporary flags.
 */
static inline void
Temporary_setFlags(Temporary this, TemporaryFlags flags)
{
  *Temporary__FLAGS(this) |= flags;
}

/*
 * Temporary_clearFlags --	Clear this Temporary flags.
 */
static inline void
Temporary_clearFlags(Temporary this, TemporaryFlags flags)
{
  *Temporary__FLAGS(this) &= ~flags;
}

/*
 * Temporary_register --	Returns register associated with Register Temporary.
 *
 * Return:	the Register, or Register__UNDEF for non-Register Temporary(ies).
 */
static inline Register
Temporary_register(const_Temporary this)
{
  Except_CHECK(Temporary_hasRegister(this) || Temporary_REGISTER(this) == Register__UNDEF);
  return Temporary_REGISTER(this);
}

/*
 * Temporary_regFile --	Returns register file associated with Register Temporary.
 *
 * Return:	the Register, or RegFile__UNDEF for non-Register Temporary(ies).
 */
static inline RegFile
Temporary_regFile(const_Temporary this)
{
  Except_CHECK(Temporary_hasRegFile(this) || Temporary_REGFILE(this) == RegFile__UNDEF);
  return Temporary_REGFILE(this);
}

/*
 * TemporaryInvariantUnion_ --	Union for the variant part of Temporary.
 */
union TemporaryInvariantUnion_ {
  void *POINTER;			// Client pointer.
  Symbol SYMBOL;			// TemporaryType_isSymbol.
  Label LABEL;				// TemporaryType_isLabel.
  Modifier MODIFIER;			// TemporaryType_isModifier.
};
typedef union TemporaryInvariantUnion_ TemporaryInvariantUnion_, *TemporaryInvariantUnion;

/*
 * TemporaryInvariant --	Extension of Temporary for Invariant Temporary(ies).
 */
struct TemporaryInvariant_ {
  Temporary_ __;
  TemporaryInvariantUnion_ UNION_;
  //@access POINTER	(TemporaryInvariant_UNION_(this).POINTER)
  //@mutate POINTER	(&TemporaryInvariant_UNION_(this).POINTER)
  //@access SYMBOL	(TemporaryInvariant_UNION_(this).SYMBOL)
  //@mutate SYMBOL	(&TemporaryInvariant_UNION_(this).SYMBOL)
  //@access LABEL	(TemporaryInvariant_UNION_(this).LABEL)
  //@mutate LABEL	(&TemporaryInvariant_UNION_(this).LABEL)
  //@access MODIFIER	(TemporaryInvariant_UNION_(this).MODIFIER)
  //@mutate MODIFIER	(&TemporaryInvariant_UNION_(this).MODIFIER)
  Relocation RELOCATION;
  ImmediateValue VALUE;
  //@access OFFSET	TemporaryInvariant_VALUE(this)
  //@mutate OFFSET	TemporaryInvariant__VALUE(this)
  //@access MEMBER	TemporaryInvariant_VALUE(this)
  //@mutate MEMBER	TemporaryInvariant__VALUE(this)
};
typedef struct TemporaryInvariant_ TemporaryInvariant_, *TemporaryInvariant;
typedef const struct TemporaryInvariant_ *const_TemporaryInvariant;
typedef struct TemporaryInvariant_ * restrict_TemporaryInvariant;

#define TemporaryInvariant___(this) ((this)->__)
#define TemporaryInvariant____(this) (&(this)->__)
#define TemporaryInvariant_UNION_(this) ((this)->UNION_)
#define TemporaryInvariant__UNION_(this) (&(this)->UNION_)
#define TemporaryInvariant_RELOCATION(this) ((this)->RELOCATION)
#define TemporaryInvariant__RELOCATION(this) (&(this)->RELOCATION)
#define TemporaryInvariant_VALUE(this) ((this)->VALUE)
#define TemporaryInvariant__VALUE(this) (&(this)->VALUE)
#define TemporaryInvariant_OFFSET(this) (TemporaryInvariant_VALUE(this))
#define TemporaryInvariant_SYMBOL(this) ((TemporaryInvariant_UNION_(this).SYMBOL))
#define TemporaryInvariant_LABEL(this) ((TemporaryInvariant_UNION_(this).LABEL))
#define TemporaryInvariant_POINTER(this) ((TemporaryInvariant_UNION_(this).POINTER))
#define TemporaryInvariant_MODIFIER(this) ((TemporaryInvariant_UNION_(this).MODIFIER))
#define TemporaryInvariant_MEMBER(this) (TemporaryInvariant_VALUE(this))
#define TemporaryInvariant__OFFSET(this) (TemporaryInvariant__VALUE(this))
#define TemporaryInvariant__SYMBOL(this) ((&TemporaryInvariant_UNION_(this).SYMBOL))
#define TemporaryInvariant__MODIFIER(this) ((&TemporaryInvariant_UNION_(this).MODIFIER))
#define TemporaryInvariant__LABEL(this) ((&TemporaryInvariant_UNION_(this).LABEL))
#define TemporaryInvariant__MEMBER(this) (TemporaryInvariant__VALUE(this))
#define TemporaryInvariant__POINTER(this) ((&TemporaryInvariant_UNION_(this).POINTER))


//
static inline Temporary
TemporaryInvariant__(TemporaryInvariant this)
{
  Temporary temporary = TemporaryInvariant____(this);
  Except_CHECK(Temporary_isInvariant(temporary));
  return temporary;
}
static inline const_Temporary
TemporaryInvariant__const(const_TemporaryInvariant this)
{
  const_Temporary temporary = TemporaryInvariant____(this);
  Except_CHECK(Temporary_isInvariant(temporary));
  return temporary;
}

/*
 * TemporaryInvariant_label --	This TemporaryInvariant Label.
 */
static inline Label
TemporaryInvariant_label(const_TemporaryInvariant this)
{
  Except_CHECK(Temporary_isInvariant(TemporaryInvariant____(this)));
  return TemporaryInvariant_LABEL(this);
}

/*
 * TemporaryInvariant_symbol --	This TemporaryInvariant Symbol.
 */
static inline Symbol
TemporaryInvariant_symbol(const_TemporaryInvariant this)
{
  Except_CHECK(Temporary_isInvariant(TemporaryInvariant____(this)));
  return TemporaryInvariant_SYMBOL(this);
}

/*
 * TemporaryInvariant_value --	This TemporaryInvariant value.
 */
static inline ImmediateValue
TemporaryInvariant_value(const_TemporaryInvariant this)
{
  Except_CHECK(Temporary_isInvariant(TemporaryInvariant____(this)));
  return TemporaryInvariant_VALUE(this);
}

/*
 * TemporaryInvariant_offset --	This TemporaryInvariant offset.
 */
static inline ImmediateValue
TemporaryInvariant_offset(const_TemporaryInvariant this)
{
  Except_CHECK(Temporary_isInvariant(TemporaryInvariant____(this)));
  return TemporaryInvariant_OFFSET(this);
}

/*
 * TemporaryInvariant_relocation --	This TemporaryInvariant relocation.
 */
static inline Relocation
TemporaryInvariant_relocation(const_TemporaryInvariant this)
{
  Except_CHECK(Temporary_isInvariant(TemporaryInvariant____(this)));
  return TemporaryInvariant_RELOCATION(this);
}

/*
 * Temporary_make --	Low-level function to make a new Temporary.
 */
static inline Temporary
Temporary_make(IBList idlist, TemporaryType type, TemporaryFlags flags)
{
  Temporary temporary = IBList_push(idlist);
  Temporary_Ctor(temporary, type, flags);
  return temporary;
}

/*
 * Temporary_setDedicated --	Force a Virtual Temporary to be dedicated.
 *
 * Note that this function only exists because copy folding may have discarded a
 * dedicated Temporary and replaced it by an Assigned Temporary. On the Open64
 * such copy folding breaks the invariant that procedure register arguments should be
 * in Dedicated Temporary(ies). Calling Temporary_setDedicated cures the problem.
 */
static inline void
Temporary_setDedicated(Temporary this)
{
  Except_CHECK(Temporary_isAssigned(this));
  *Temporary__TYPE(this) = TemporaryType_Dedicated;
}

/*
 * Temporary_compare --	Compare two Temporary(ies) for total ordering.
 */
int
Temporary_compare(const_Temporary this, const_Temporary that);

/*
 * Temporary_pretty --	Pretty print of this Temporary.
 */
bool
Temporary_pretty(const_Temporary this, FILE *file);

/*
 * Temporary_isLifetime --	True if the lifetime of this Temporary must be considered.
 */
static inline bool
Temporary_isLifetime(Temporary this)
{
  Convention convention = Convention__DEFAULT;	// FIXME!
  return    Temporary_isVirtual(this)
         || (   Temporary_hasRegFile(this)
             && Register_isLifetime(Temporary_REGISTER(this), convention));
}

/*
 * Temporary_assignRegister --	Assign a Register to this Temporary.
 */
void
Temporary_assignRegister(Temporary this, Register registre);

/*
 * Temporary_unassignRegister --	Un-assign this Temporary.
 */
void
Temporary_unassignRegister(Temporary this);

/*
 * Temporary_setWidth --	Set the bit-width of this Register Temporary.
 */
static inline void
Temporary_setWidth(Temporary this, unsigned width)
{
  *Temporary__WIDTH(this) = width;
}

/*
 * TemporaryStack --	Stack of Temporary(ies).
 */
struct TemporaryStack_ {
  //@args	Memory memory, int maxCount
  PtrSeq_ __;		// Underlying PtrSeq.
};
typedef struct TemporaryStack_ TemporaryStack_, *TemporaryStack;
typedef const struct TemporaryStack_ *const_TemporaryStack;
typedef struct TemporaryStack_ * restrict_TemporaryStack;

extern TemporaryStack
TemporaryStack_Ctor(TemporaryStack this, Memory memory, int maxCount);

extern TemporaryStack
TemporaryStack_Copy(TemporaryStack this, const_TemporaryStack that);

extern void
TemporaryStack_Dtor(TemporaryStack this);

extern size_t
TemporaryStack_Size(Memory memory, int maxCount);

#define TemporaryStack___(this) ((this)->__)
#define TemporaryStack____(this) (&(this)->__)


//
static inline PtrSeq
TemporaryStack__(TemporaryStack this)
{
  return TemporaryStack____(this);
}
static inline const_PtrSeq
TemporaryStack__const(const_TemporaryStack this)
{
  return TemporaryStack____(this);
}

/*
 * TemporaryStack_FOREACH_Temporary --	Iterate over this TemporaryStack Temporary(ies).
 */
#define TemporaryStack_FOREACH_Temporary(this, temporary) { \
  const_PtrSeq TemporaryStack____ = TemporaryStack__const(this); \
  PtrSeq_FOREACH(TemporaryStack____, Temporary, temporary) {
#define TemporaryStack_ENDEACH_Temporary \
  } PtrSeq_ENDEACH \
}

/*
 * TemporaryStack_FORBACK_Temporary --	Iterate over this TemporaryStack Temporary(ies).
 */
#define TemporaryStack_FORBACK_Temporary(this, temporary) { \
  const_PtrSeq TemporaryStack____ = TemporaryStack__const(this); \
  PtrSeq_FORBACK(TemporaryStack____, Temporary, temporary) {
#define TemporaryStack_ENDBACK_Temporary \
  } PtrSeq_ENDBACK \
}

//
#define TemporaryStack_resize(this, maxCount) \
  PtrSeq_resize(TemporaryStack__(this), maxCount)
#define TemporaryStack_empty(this) \
  PtrSeq_empty(TemporaryStack__(this))
#define TemporaryStack_memory(this) \
  PtrSeq_memory(TemporaryStack__const(this))
#define TemporaryStack_base_(this) (Temporary *) \
  PtrSeq_base_(TemporaryStack__(this))
#define TemporaryStack_base(this) (const Temporary *) \
  PtrSeq_base(TemporaryStack__const(this))
#define TemporaryStack_past(this) (const Temporary *) \
  PtrSeq_past(TemporaryStack__const(this))
#define TemporaryStack_isEmpty(this) \
  PtrSeq_isEmpty(TemporaryStack__const(this))
#define TemporaryStack_isSingle(this) \
  PtrSeq_isSingle(TemporaryStack__const(this))
#define TemporaryStack_isFull(this) \
  PtrSeq_isFull(TemporaryStack__const(this))
#define TemporaryStack_count(this) \
  PtrSeq_count(TemporaryStack__const(this))
#define TemporaryStack_access(this, index) (Temporary) \
  PtrSeq_access(TemporaryStack__const(this), index)
#define TemporaryStack_bottom(this) (Temporary) \
  PtrSeq_bottom(TemporaryStack__const(this))
#define TemporaryStack_top(this) (Temporary) \
  PtrSeq_top(TemporaryStack__const(this))
#define TemporaryStack_pop(this) \
  PtrSeq_pop(TemporaryStack__(this))
#define TemporaryStack_push(this, temporary) \
  PtrSeq_push(TemporaryStack__(this), temporary)
#define TemporaryStack_push2(this, temporary) \
  PtrSeq_push2(TemporaryStack__(this), temporary)

/*
 * TemporarySet --	Set of Temporary(ies).
 */
struct TemporarySet_ {
  //@args	Memory memory, int32_t estimate
  PtrSet_ __;		// The PtrSet_ that implements the set.
  //@access MEMORY	PtrSet_memory(TemporarySet____(this))
};
typedef struct TemporarySet_ TemporarySet_, *TemporarySet;
typedef const struct TemporarySet_ *const_TemporarySet;
typedef struct TemporarySet_ * restrict_TemporarySet;

extern TemporarySet
TemporarySet_Ctor(TemporarySet this, Memory memory, int32_t estimate);

extern TemporarySet
TemporarySet_Copy(TemporarySet this, const_TemporarySet that);

extern void
TemporarySet_Dtor(TemporarySet this);

extern size_t
TemporarySet_Size(Memory memory, int32_t estimate);

#define TemporarySet___(this) ((this)->__)
#define TemporarySet____(this) (&(this)->__)
#define TemporarySet_MEMORY(this) (PtrSet_memory(TemporarySet____(this)))


//
static inline PtrSet
TemporarySet__(TemporarySet this)
{
  return TemporarySet____(this);
}
static inline const_PtrSet
TemporarySet__const(const_TemporarySet this)
{
  return TemporarySet____(this);
}

/*
 * TemporarySet_FOREACH_Temporary --	Iterate over this TemporarySet Temporary(ies).
 */
#define TemporarySet_FOREACH_Temporary(this, temporary) { \
  const_PtrSet TemporarySet____ = TemporarySet__const(this); \
  PtrSet_FOREACH(TemporarySet____, Temporary, temporary) {
#define TemporarySet_ENDEACH_Temporary \
  } PtrSet_ENDEACH \
}

/*
 * TemporarySet_FORSORT_Temporary --	Iterate over this TemporarySet sorted Temporary(ies).
 */
#define TemporarySet_FORSORT_Temporary(this, compare, temporary) { \
  const_PtrSet TemporarySet____ = TemporarySet__const(this); \
  PtrSet_FORSORT(TemporarySet____, compare, Temporary, temporary) {
#define TemporarySet_ENDSORT_Temporary \
  } PtrSet_ENDSORT \
}

//
#define TemporarySet_memory(this) \
  PtrSet_memory(TemporarySet__const(this))
#define TemporarySet_base_(this) (Temporary *) \
  PtrSet_base_(TemporarySet__(this))
#define TemporarySet_base(this) (const Temporary *) \
  PtrSet_base(TemporarySet__const(this))
#define TemporarySet_past(this) (const Temporary *) \
  PtrSet_past(TemporarySet__const(this))
#define TemporarySet_isOrdered(this) \
  PtrSet_isOrdered(TemporarySet__const(this))
#define TemporarySet_count(this) \
  PtrSet_count(TemporarySet__const(this))
#define TemporarySet_isEmpty(this) \
  PtrSet_isEmpty(TemporarySet__const(this))
#define TemporarySet_isSingle(this) \
  PtrSet_isSingle(TemporarySet__const(this))
#define TemporarySet_empty(this) \
  PtrSet_empty(TemporarySet__(this))
#define TemporarySet_clearOrdered(this) \
  PtrSet_clearOrdered(TemporarySet__(this))
#define TemporarySet_addMember(this, temporary)	\
  PtrSet_addMember(TemporarySet__(this), (PtrSetMember)temporary)
#define TemporarySet_makeOrder(this) \
  PtrSet_makeOrder(TemporarySet__(this))
#define TemporarySet_contains(this, temporary)	\
  PtrSet_contains(TemporarySet__const(this), (PtrSetMember)temporary)
#define TemporarySet_insert(this, temporary)	\
  PtrSet_insert(TemporarySet__(this), (PtrSetMember)temporary)
#define TemporarySet_remove(this, temporary)	\
  PtrSet_remove(TemporarySet__(this), (PtrSetMember)temporary)
#define TemporarySet_equals(this, that)	\
  PtrSet_equals(TemporarySet__const(this), TemporarySet__const(that))
#define TemporarySet_assign(this, that)	\
  PtrSet_assign(TemporarySet__(this), TemporarySet__const(that))
#define TemporarySet_union(this, that)	\
  PtrSet_union(TemporarySet__(this), TemporarySet__const(that))
#define TemporarySet_inter(this, that)	\
  PtrSet_inter(TemporarySet__(this), TemporarySet__const(that))
#define TemporarySet_diff(this, that)	\
  PtrSet_diff(TemporarySet__(this), TemporarySet__const(that))
#define TemporarySet_comprises(this, that)	\
  PtrSet_comprises(TemporarySet__const(this), TemporarySet__const(that))
#define TemporarySet_overlap(this, that)	\
  PtrSet_overlap(TemporarySet__const(this), TemporarySet__const(that))
#define TemporarySet_choose(this) \
  PtrSet_choose(TemporarySet__(this))

//
TemporarySet
TemporarySet_make(Memory memory, int32_t estimate);

//
TemporarySet
TemporarySet_kill(TemporarySet this);

/*
 * TemporarySet_pretty --	Pretty-print this TemporarySet.
 */
bool
TemporarySet_pretty(TemporarySet this, FILE *file);

/*
 * TemporarySet_display --	Display this TemporarySet in VCG format.
 */
bool
TemporarySet_display(TemporarySet this, FILE *file);

/*
 * TemporaryTable --	Specialization of IndexedTable for Temporary(ies).
 */
struct TemporaryTable_ {
  //@args	Memory memory, int32_t estimate
  IndexedTable_ __;
  //@access MEMORY	IndexedTable_memory(TemporaryTable____(this))
  //@access COUNT	IndexedTable_count(TemporaryTable____(this))
  //@access BASE	(const Temporary *)IndexedTable_base(TemporaryTable____(this))
};
typedef struct TemporaryTable_ TemporaryTable_, *TemporaryTable;
typedef const struct TemporaryTable_ *const_TemporaryTable;
typedef struct TemporaryTable_ * restrict_TemporaryTable;

extern TemporaryTable
TemporaryTable_Ctor(TemporaryTable this, Memory memory, int32_t estimate);

extern TemporaryTable
TemporaryTable_Copy(TemporaryTable this, const_TemporaryTable that);

extern void
TemporaryTable_Dtor(TemporaryTable this);

extern size_t
TemporaryTable_Size(Memory memory, int32_t estimate);

#define TemporaryTable___(this) ((this)->__)
#define TemporaryTable____(this) (&(this)->__)
#define TemporaryTable_BASE(this) ((const Temporary *)IndexedTable_base(TemporaryTable____(this)))
#define TemporaryTable_COUNT(this) (IndexedTable_count(TemporaryTable____(this)))
#define TemporaryTable_MEMORY(this) (IndexedTable_memory(TemporaryTable____(this)))


//
static inline IndexedTable
TemporaryTable__(TemporaryTable this)
{
  return TemporaryTable____(this);
}
static inline const_IndexedTable
TemporaryTable__const(const_TemporaryTable this)
{
  return TemporaryTable____(this);
}

//
#define TemporaryTable_memory(this) \
  IndexedTable_memory(TemporaryTable__const(this))
#define TemporaryTable_count(this) \
  IndexedTable_count(TemporaryTable__const(this))
#define TemporaryTable_empty(this) \
  IndexedTable_empty(TemporaryTable__(this))
#define TemporaryTable_contains(this, temporary) \
  IndexedTable_contains(TemporaryTable__const(this), Temporary__const(temporary))
#define TemporaryTable_insert(this, temporary) \
  IndexedTable_insert(TemporaryTable__(this), Temporary__(temporary))
#define TemporaryTable_remove(this, temporary) \
  IndexedTable_remove(TemporaryTable__(this), Temporary__const(temporary))
#define TemporaryTable_removeHigh(this, temporary, high) \
  IndexedTable_removeHigh(TemporaryTable__(this), Temporary__const(temporary), high)

//
static inline const Temporary *
TemporaryTable_base(const_TemporaryTable this)
{
  return TemporaryTable_BASE(this);
}

/*
 * TemporaryTable_FOREACH_Temporary --	Iterate this TemporaryTable Temporary(ies).
 */
#define TemporaryTable_FOREACH_Temporary(this, temporary) { \
  IndexedTable_FOREACH_Indexed(TemporaryTable__const(this), \
                                   Temporary, temporary) { \
    int32_t TemporaryTable_Temporary_INDEX = IndexedTable_INDEX;
#define TemporaryTable_ENDEACH_Temporary \
  } IndexedTable_ENDEACH_Indexed; \
}

//
bool
TemporaryTable_pretty(const_TemporaryTable this, FILE *file);

/*
 * TemporarySparse --	Sparse set of Temporary(ies) based on INDEX Sparse set.
 */
struct TemporarySparse_ {
  //@args	Memory memory, TemporaryTable table
  IndexedSparse_ __;
  //@access MEMORY	IndexedSparse_memory(TemporarySparse____(this))
  //@access BASE	(const Temporary *)IndexedSparse_base(TemporarySparse____(this))
};
typedef struct TemporarySparse_ TemporarySparse_, *TemporarySparse;
typedef const struct TemporarySparse_ *const_TemporarySparse;
typedef struct TemporarySparse_ * restrict_TemporarySparse;

extern TemporarySparse
TemporarySparse_Ctor(TemporarySparse this, Memory memory, TemporaryTable table);

extern TemporarySparse
TemporarySparse_Copy(TemporarySparse this, const_TemporarySparse that);

extern void
TemporarySparse_Dtor(TemporarySparse this);

extern size_t
TemporarySparse_Size(Memory memory, TemporaryTable table);

#define TemporarySparse___(this) ((this)->__)
#define TemporarySparse____(this) (&(this)->__)
#define TemporarySparse_BASE(this) ((const Temporary *)IndexedSparse_base(TemporarySparse____(this)))
#define TemporarySparse_MEMORY(this) (IndexedSparse_memory(TemporarySparse____(this)))


//
static inline IndexedSparse
TemporarySparse__(TemporarySparse this)
{
  return TemporarySparse____(this);
}
static inline const_IndexedSparse
TemporarySparse__const(const_TemporarySparse this)
{
  return TemporarySparse____(this);
}

//
#define TemporarySparse_empty(this) \
  IndexedSparse_empty(TemporarySparse__(this))
#define TemporarySparse_memory(this) \
  IndexedSparse_memory(TemporarySparse__const(this))
#define TemporarySparse_isEmpty(this) \
  IndexedSparse_isEmpty(TemporarySparse__const(this))
#define TemporarySparse_isSingle(this) \
  IndexedSparse_isSingle(TemporarySparse__const(this))
#define TemporarySparse_count(this) \
  IndexedSparse_count(TemporarySparse__const(this))
#define TemporarySparse_choose(this) \
  IndexedSparse_choose(TemporarySparse__(this))
#define TemporarySparse_contains(this, temporary) \
  IndexedSparse_contains(TemporarySparse__const(this), Temporary__const(temporary))
#define TemporarySparse_insert(this, temporary) \
  IndexedSparse_insert(TemporarySparse__(this), Temporary__const(temporary))
#define TemporarySparse_remove(this, temporary) \
  IndexedSparse_remove(TemporarySparse__(this), Temporary__const(temporary))
#define TemporarySparse_equals(this, that) \
  IndexedSparse_equals(TemporarySparse__const(this), TemporarySparse__const(that))
#define TemporarySparse_assign(this, that) \
  IndexedSparse_assign(TemporarySparse__(this), TemporarySparse__const(that))
#define TemporarySparse_union(this, that) \
  IndexedSparse_union(TemporarySparse__(this), TemporarySparse__const(that))
#define TemporarySparse_inter(this, that) \
  IndexedSparse_inter(TemporarySparse__(this), TemporarySparse__const(that))
#define TemporarySparse_diff(this, that) \
  IndexedSparse_diff(TemporarySparse__(this), TemporarySparse__const(that))
#define TemporarySparse_position(this, temporary) \
  IndexedSparse_position(TemporarySparse__const(this), Temporary__const(temporary))

/*
 * TemporarySparse_FOREACH_Temporary --	Iterate this TemporarySparse Temporary(ies).
 */
#define TemporarySparse_FOREACH_Temporary(this, temporary) { \
  IndexedSparse_FOREACH_Indexed(TemporarySparse__const(this), Temporary, temporary) { \
    int32_t TemporarySparse_POSITION = IndexedSparse_POSITION; \
    int32_t TemporarySparse_INDEX = IndexedSparse_INDEX;
#define TemporarySparse_ENDEACH_Temporary \
  } IndexedSparse_ENDEACH_Indexed; \
}

//
static inline TemporaryTable
TemporarySparse_table(TemporarySparse this)
{
  return (TemporaryTable)IndexedSparse_table(TemporarySparse____(this));
}
static inline const_TemporaryTable
TemporarySparse_table_const(const_TemporarySparse this)
{
  return (const_TemporaryTable)IndexedSparse_table_const(TemporarySparse____(this));
}

//
void
TemporarySparse_insertSet(TemporarySparse this, const_TemporarySet temporarySet);

//
bool
TemporarySparse_pretty(const_TemporarySparse this, FILE *file);

/*
 * TemporaryMap --	Map Temporary to user-defined object.
 *
 * This map is based on the Temporary_INDEX fields, which must be properly
 * initialized for instance by entering the Temporary(ies) in a TemporaryTable
 * before the TemporaryMap is constructed.
 */
struct TemporaryMap_ {
  //@args	Memory memory, int32_t maxCount, size_t itemSize
  Memory MEMORY;
  int32_t MAXCOUNT;
  size_t ITEMSIZE;
  void *BASE;
};
typedef struct TemporaryMap_ TemporaryMap_, *TemporaryMap;
typedef const struct TemporaryMap_ *const_TemporaryMap;
typedef struct TemporaryMap_ * restrict_TemporaryMap;

extern TemporaryMap
TemporaryMap_Ctor(TemporaryMap this, Memory memory, int32_t maxCount, size_t itemSize);

extern TemporaryMap
TemporaryMap_Copy(TemporaryMap this, const_TemporaryMap that);

extern void
TemporaryMap_Dtor(TemporaryMap this);

extern size_t
TemporaryMap_Size(Memory memory, int32_t maxCount, size_t itemSize);

#define TemporaryMap_MEMORY(this) ((this)->MEMORY)
#define TemporaryMap__MEMORY(this) (&(this)->MEMORY)
#define TemporaryMap_MAXCOUNT(this) ((this)->MAXCOUNT)
#define TemporaryMap__MAXCOUNT(this) (&(this)->MAXCOUNT)
#define TemporaryMap_ITEMSIZE(this) ((this)->ITEMSIZE)
#define TemporaryMap__ITEMSIZE(this) (&(this)->ITEMSIZE)
#define TemporaryMap_BASE(this) ((this)->BASE)
#define TemporaryMap__BASE(this) (&(this)->BASE)


/*
 * TemporaryMap_resize --	Resize this TemporaryMap with a new maxCount.
 */
void
TemporaryMap_resize(TemporaryMap this, int32_t maxCount);

/*
 * TemporaryMap_resize --	Resize this TemporaryMap memory.
 */
void
TemporaryMap_release(TemporaryMap this, void (*release)(void * item));

/*
 * TemporaryMap_base --	For use in TemporaryMap_FOREACH.
 */
static inline void *
TemporaryMap_base(TemporaryMap this)
{
  return TemporaryMap_BASE(this);
}

/*
 * TemporaryMap_past --	For use in TemporaryMap_FOREACH.
 */
static inline void *
TemporaryMap_past(TemporaryMap this)
{
  char *base = TemporaryMap_BASE(this);
  int32_t maxCount = TemporaryMap_MAXCOUNT(this);
  size_t itemSize = TemporaryMap_ITEMSIZE(this);
  return base + maxCount*itemSize;
}

/*
 * TemporaryMap_itemSize --	For use in TemporaryMap_FOREACH.
 */
static inline size_t
TemporaryMap_itemSize(TemporaryMap this)
{
  return TemporaryMap_ITEMSIZE(this);
}

/*
 * TemporaryMap_FOREACH --	Iterate over this TemporaryMap objects.
 */
#define TemporaryMap_FOREACH(this, Type, iter) { \
  size_t TemporaryMap_ITEMSIZE = TemporaryMap_itemSize(this); \
  Type *TemporaryMap_BASE = (Type *)TemporaryMap_base(this); \
  Type *TemporaryMap_PAST = (Type *)TemporaryMap_past(this), *(iter); \
  for (iter = TemporaryMap_BASE; iter < TemporaryMap_PAST; ++(iter)) { \
    int32_t TemporaryMap_INDEX = iter - TemporaryMap_BASE;
#define TemporaryMap_ENDEACH \
  } \
}

/*
 * TemporaryMap_access --	Access the object mapped by the temporary.
 */
void *
TemporaryMap_access(TemporaryMap this, Temporary temporary);

//
bool
Temporary_DTD(FILE *file);

//
bool
Temporary_dumpXML(Temporary this, XMLGen gen);


#endif//Temporary_h_INCLUDED
