#ifndef Lifetime__h_INCLUDED
#define Lifetime__h_INCLUDED

//
#include "XFA/Lifetime.h"

/*
 * Lifetime__firstEffect --	The first Effect pointer of this Lifetime.
 */
static inline Effect *
Lifetime__firstEffect(Lifetime this)
{
  int32_t firstIndex = Lifetime_FIRSTINDEX(this);
  return EffectSeq_base(Lifetime_EFFECTSEQ(this)) + firstIndex;
}
#define Lifetime_firstEffect(this)	(*Lifetime__firstEffect(this))

/*
 * Lifetime__lastEffect --	The last Effect pointer of this Lifetime.
 */
static inline Effect *
Lifetime__lastEffect(Lifetime this)
{
  int32_t lastIndex = Lifetime_LASTINDEX(this);
  return EffectSeq_base(Lifetime_EFFECTSEQ(this)) + lastIndex;
}
#define Lifetime_lastEffect(this)	(*Lifetime__lastEffect(this))

/*
 * Lifetime_mayIgnore --	Find if this Lifetime may be ignored.
 *
 * We may ignore the Lifetime(s) that only have clobber or side effects.
 */
bool
Lifetime_mayIgnore(Lifetime this);

/*
 * Lifetime_compare --	Compare function for sorting the Lifetime(s).
 */
int
Lifetime_compare(const void *this_pointer, const void *that_pointer);

/*
 * Lifetime_complete --	Complete this Lifetime.
 *
 * Completion rotates the effects of this Lifetime so the starting effect is a
 * Def (Write) if there is any. For all the Lifetimes that are not on register with
 * renaming, the first Def is replicated at the end of this Lifetime effects.
 */
void
Lifetime_complete(Lifetime this, Optimize optimize, Configure configure);

/*
 * Lifetime_analyze --	Analyze this Lifetime and set attributes.
 */
void
Lifetime_analyze(Lifetime this, Optimize optimize, Configure configure);

/*
 * Lifetime_setSimpleInductionFlags --	Set Lifetime and Operation flags for simple
 * inductions.
 *
 * A simple induction is detected even in case of block scheduling. However in
 * such cases only the uses after the induction update will be relaxable. This is
 * because the Label operation introduces a fake definition in the Lifetime.
 */
void
Lifetime_setSimpleInductionFlags(Lifetime this, Optimize optimize, Configure configure);

/*
 * Lifetime_setPreloading --	Set Operation_isPreload
 */
void
Lifetime_setPreloading(Lifetime this, Optimize optimize, Configure configure);

/*
 * Lifetime_linkEffects --	Link this Lifetime Effect PRED and SUCC fields for
 * later use by dependence graph construction.
 *
 * The linking of effects crates a poor man's Def-Use and Use-Def chains within a
 * Lifetime (which is local to a BlockTrace). The Defs (Write) are double-linked
 * together. The Uses (non-Write) are linked to the previous and the next Defs.
 */
void
Lifetime_linkEffects(Lifetime this);

/*
 * Lifetime_relinkEffects --	Relink this Lifetime Effect PRED and SUCC
 * fields according to the supplied ddgNodes passed in.
 *
 * The supplied operation set contains the tails and the head operation of the
 * dependences that are already accounted for, such as the externally supplied
 * memory dependences. The purpose of Lifetime_relinkEffects is to reroute the
 * PRED and SUCC chains around the supplied operations.
 */
void
Lifetime_relinkEffects(Lifetime this, PtrSet ddgNodes);

/*
 * Lifetime_computeLifespan --	Compute and return this Lifetime LIFESPAN.
 *
 * The lifespan is assumed to start at the earliest start date, and stop at the
 * latest start date in this Lifetime. TODO: correct the lifespan value depending
 * on the operand read and write time offsets.
 */
int
Lifetime_computeLifespan(Lifetime this, int16_t lambda, bool isAcyclic);

/*
 * Lifetime_fillRenaming --	Make the renaming Temporaries needed by modulo expansion.
 *
 * Require that this Lifetime be inlined within an IDList sorted with Lifetime_compare.
 */
void
Lifetime_fillRenaming(Lifetime this, Procedure procedure, int lambda, int kunroll);

/*
 * Lifetime_findReachingLifetime --	Find the Lifetime that reaches an operation argument.
 *
 * Require that this Lifetime be inlined within an IDList sorted with Lifetime_compare.
 */
Lifetime
Lifetime_findReachingLifetime(Lifetime this, Operation operation, int argIndex);

//
struct LifetimeTable_ {
  IDList_ LIFETIMES_;		//< The IDList where the Lifetime(s) are inlined.
  //@access LIFETIMES	LifetimeTable__LIFETIMES_(this)
  //@access MEMORY	IDList_memory(LifetimeTable_LIFETIMES(this))
  //@access COUNT	IDList_count(LifetimeTable_LIFETIMES(this))
  //@access FIRST	(Lifetime)(IDList_firstItem(LifetimeTable_LIFETIMES(this)))
  //@access LAST	(Lifetime)(IDList_lastItem(LifetimeTable_LIFETIMES(this)))
  TemporaryMap_ TEMPORARYMAP_;	//< Map from Temporary INDEX to Lifetime.
  //@access TEMPORARYMAP	LifetimeTable__TEMPORARYMAP_(this)
  Optimize OPTIMIZE;
  Configure CONFIGURE;
};

extern LifetimeTable
LifetimeTable_Ctor(LifetimeTable this, Memory memory, Optimize optimize, Configure configure);

extern LifetimeTable
LifetimeTable_Copy(LifetimeTable this, const_LifetimeTable that);

extern void
LifetimeTable_Dtor(LifetimeTable this);

extern size_t
LifetimeTable_Size(Memory memory, Optimize optimize, Configure configure);

#define LifetimeTable_LIFETIMES_(this) ((this)->LIFETIMES_)
#define LifetimeTable__LIFETIMES_(this) (&(this)->LIFETIMES_)
#define LifetimeTable_TEMPORARYMAP_(this) ((this)->TEMPORARYMAP_)
#define LifetimeTable__TEMPORARYMAP_(this) (&(this)->TEMPORARYMAP_)
#define LifetimeTable_OPTIMIZE(this) ((this)->OPTIMIZE)
#define LifetimeTable__OPTIMIZE(this) (&(this)->OPTIMIZE)
#define LifetimeTable_CONFIGURE(this) ((this)->CONFIGURE)
#define LifetimeTable__CONFIGURE(this) (&(this)->CONFIGURE)
#define LifetimeTable_LIFETIMES(this) (LifetimeTable__LIFETIMES_(this))
#define LifetimeTable_FIRST(this) ((Lifetime)(IDList_firstItem(LifetimeTable_LIFETIMES(this))))
#define LifetimeTable_TEMPORARYMAP(this) (LifetimeTable__TEMPORARYMAP_(this))
#define LifetimeTable_COUNT(this) (IDList_count(LifetimeTable_LIFETIMES(this)))
#define LifetimeTable_MEMORY(this) (IDList_memory(LifetimeTable_LIFETIMES(this)))
#define LifetimeTable_LAST(this) ((Lifetime)(IDList_lastItem(LifetimeTable_LIFETIMES(this))))


/*
 * LifetimeTable_first --	First lifetime or NULL.
 */
static inline Lifetime
LifetimeTable_first(LifetimeTable this)
{
  return LifetimeTable_FIRST(this);
}

/*
 * LifetimeTable_last --	Last lifetime or NULL.
 */
static inline Lifetime
LifetimeTable_last(LifetimeTable this)
{
  return LifetimeTable_LAST(this);
}

//
static inline bool
LifetimeTable_isEmpty(LifetimeTable this)
{
  return IDList_isEmpty(LifetimeTable_LIFETIMES(this));
}

//
static inline int32_t
LifetimeTable_count(LifetimeTable this)
{
  return IDList_count(LifetimeTable_LIFETIMES(this));
}

/*
 * LifetimeTable_grow --	Grow this LifetimeTable from =effectSeq=.
 */
void
LifetimeTable_grow(LifetimeTable this, Temporary temporary,
                   EffectSeq effectSeq, unsigned lifetimeFlags);

/*
 * LifetimeTable_build --	Build this LifetimeTable from an EffectTable.
 */
void
LifetimeTable_build(LifetimeTable this, EffectTable effectTable,
                    bool isInnerLoop, bool isPrePass);

/*
 * LifetimeTable_linkEffects --	Link the LifetimeTable Effect(s).
 */
void
LifetimeTable_linkEffects(LifetimeTable this, PtrSet ddgNodes);

/*
 * LifetimeTable_refineDependences --	Refine the memory dependences by analysing addresses.
 */
void
LifetimeTable_refineDependences(LifetimeTable this, PtrSet ddgNodes,
                                const_DependenceTable dependenceTable);

/*
 * LifetimeTable_computeMaxSpan --	Compute the maximum Lifetime span.
 */
int
LifetimeTable_computeMaxSpan(LifetimeTable this, int lambda, bool isAcyclic);

/*
 * LifetimeTable_fillRenaming --	Fill the renaming stacks.
 */
void
LifetimeTable_fillRenaming(LifetimeTable this, Procedure procedure,
                           int lambda, int kunroll);

/*
 * Lifetime_INIT --	Initialize the Lifetime module.
 */
void
Lifetime_INIT(void);

/*
 * Lifetime_FINI --	Finalize the Lifetime module.
 */
void
Lifetime_FINI(void);


#endif//Lifetime__h_INCLUDED
